@@ -222,10 +222,9 @@ function WindowList() {
 WindowList.prototype = {
     _init : function () {
         this.actor = new St.BoxLayout({ name: 'Windows', vertical: true, style: 'spacing: 8px' });
-        let display = global.screen.get_display();
         let tracker = Shell.WindowTracker.get_default();
         this._updateId = Main.initializeDeferredWork(this.actor, Lang.bind(this, this._updateWindowList));
-        display.connect('window-created', Lang.bind(this, this._updateWindowList));
+        global.display.connect('window-created', Lang.bind(this, this._updateWindowList));
         tracker.connect('tracked-windows-changed', Lang.bind(this, this._updateWindowList));
     },
 
@@ -248,7 +247,7 @@ WindowList.prototype = {
             box.add(propsBox);
             propsBox.add(new St.Label({ text: 'wmclass: ' + metaWindow.get_wm_class() }));
             let app = tracker.get_window_app(metaWindow);
-            if (app != null && !app.is_transient()) {
+            if (app != null && !app.is_window_backed()) {
                 let icon = app.create_icon_texture(22);
                 let propBox = new St.BoxLayout({ style: 'spacing: 6px; ' });
                 propsBox.add(propBox);
@@ -639,23 +638,32 @@ Extensions.prototype = {
                                         name: 'lookingGlassExtensions' });
         this._noExtensions = new St.Label({ style_class: 'lg-extensions-none',
                                              text: _("No extensions installed") });
+        this._numExtensions = 0;
         this._extensionsList = new St.BoxLayout({ vertical: true,
                                                   style_class: 'lg-extensions-list' });
+        this._extensionsList.add(this._noExtensions);
         this.actor.add(this._extensionsList);
-        this._loadExtensionList();
+
+        for (let uuid in ExtensionSystem.extensionMeta)
+            this._loadExtension(null, uuid);
+
+        ExtensionSystem.connect('extension-loaded',
+                                Lang.bind(this, this._loadExtension));
     },
 
-    _loadExtensionList: function() {
-        let extensions = ExtensionSystem.extensionMeta;
-        let totalExtensions = 0;
-        for (let uuid in extensions) {
-            let extensionDisplay = this._createExtensionDisplay(extensions[uuid]);
-            this._extensionsList.add(extensionDisplay);
-            totalExtensions++;
-        }
-        if (totalExtensions == 0) {
-            this._extensionsList.add(this._noExtensions);
-        }
+    _loadExtension: function(o, uuid) {
+        let extension = ExtensionSystem.extensionMeta[uuid];
+        // There can be cases where we create dummy extension metadata
+        // that's not really a proper extension. Don't bother with these.
+        if (!extension.name)
+            return;
+
+        let extensionDisplay = this._createExtensionDisplay(extension);
+        if (this._numExtensions == 0)
+            this._extensionsList.remove_actor(this._noExtensions);
+
+        this._numExtensions ++;
+        this._extensionsList.add(extensionDisplay);
     },
 
     _onViewSource: function (actor) {
@@ -682,6 +690,8 @@ Extensions.prototype = {
                 return _("Error");
             case ExtensionSystem.ExtensionState.OUT_OF_DATE:
                 return _("Out of date");
+            case ExtensionSystem.ExtensionState.DOWNLOADING:
+                return _("Downloading");
         }
         return 'Unknown'; // Not translated, shouldn't appear
     },
@@ -692,7 +702,7 @@ Extensions.prototype = {
                                    text: meta.name });
         box.add(name, { expand: true });
         let description = new St.Label({ style_class: 'lg-extension-description',
-                                         text: meta.description });
+                                         text: meta.description || 'No description' });
         box.add(description, { expand: true });
 
         let metaBox = new St.BoxLayout();
@@ -751,7 +761,9 @@ LookingGlass.prototype = {
                                         Lang.bind(this, this._updateFont));
         this._updateFont();
 
-        Main.uiGroup.add_actor(this.actor);
+        // we add it to the chrome because we want it to appear to slide
+        // out from underneath the panel
+        Main.layoutManager.addChrome(this.actor);
 
         this._objInspector = new ObjInspector();
         Main.uiGroup.add_actor(this._objInspector.actor);
@@ -962,7 +974,7 @@ LookingGlass.prototype = {
 
         this._notebook.selectIndex(0);
         this.actor.show();
-        this.actor.lower(Main.chrome.actor);
+        this.actor.lower_bottom();
         this._open = true;
         this._history.lastItem();
 
@@ -993,6 +1005,7 @@ LookingGlass.prototype = {
 
         Main.popModal(this._entry);
 
+        this.actor.lower_bottom();
         Tweener.addTween(this.actor, { time: 0.5 / St.get_slow_down_factor(),
                                        transition: 'easeOutQuad',
                                        y: this._hiddenY,
