@@ -12,18 +12,22 @@ const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const St = imports.gi.St;
 
-const Chrome = imports.ui.chrome;
+const AutomountManager = imports.ui.automountManager;
+const AutorunManager = imports.ui.autorunManager;
 const CtrlAltTab = imports.ui.ctrlAltTab;
 const EndSessionDialog = imports.ui.endSessionDialog;
 const PolkitAuthenticationAgent = imports.ui.polkitAuthenticationAgent;
 const Environment = imports.ui.environment;
 const ExtensionSystem = imports.ui.extensionSystem;
+const Keyboard = imports.ui.keyboard;
 const MessageTray = imports.ui.messageTray;
 const Overview = imports.ui.overview;
 const Panel = imports.ui.panel;
 const PlaceDisplay = imports.ui.placeDisplay;
 const RunDialog = imports.ui.runDialog;
+const Layout = imports.ui.layout;
 const LookingGlass = imports.ui.lookingGlass;
+const NetworkAgent = imports.ui.networkAgent;
 const NotificationDaemon = imports.ui.notificationDaemon;
 const WindowAttentionHandler = imports.ui.windowAttentionHandler;
 const Scripting = imports.ui.scripting;
@@ -38,7 +42,8 @@ const Util = imports.misc.util;
 const DEFAULT_BACKGROUND_COLOR = new Clutter.Color();
 DEFAULT_BACKGROUND_COLOR.from_pixel(0x2266bbff);
 
-let chrome = null;
+let automountManager = null;
+let autorunManager = null;
 let panel = null;
 let hotCorners = [];
 let placesManager = null;
@@ -59,13 +64,94 @@ let uiGroup = null;
 let magnifier = null;
 let xdndHandler = null;
 let statusIconDispatcher = null;
+let keyboard = null;
+let layoutManager = null;
+let networkAgent = null;
 let _errorLogStack = [];
 let _startDate;
 let _defaultCssStylesheet = null;
 let _cssStylesheet = null;
+let _gdmCssStylesheet = null;
 
 let background = null;
 
+function _createUserSession() {
+    // Load the calendar server. Note that we are careful about
+    // not loading any events until the user presses the clock
+    global.launch_calendar_server();
+
+    placesManager = new PlaceDisplay.PlacesManager();
+    telepathyClient = new TelepathyClient.Client();
+    automountManager = new AutomountManager.AutomountManager();
+    autorunManager = new AutorunManager.AutorunManager();
+    networkAgent = new NetworkAgent.NetworkAgent();
+}
+
+function _createGDMSession() {
+    // We do this this here instead of at the top to prevent GDM
+    // related code from getting loaded in normal user sessions
+    const LoginDialog = imports.gdm.loginDialog;
+
+    let loginDialog = new LoginDialog.LoginDialog();
+    loginDialog.connect('loaded', function() {
+                            loginDialog.open();
+                        });
+}
+
+function _initRecorder() {
+    let recorderSettings = new Gio.Settings({ schema: 'org.gnome.shell.recorder' });
+
+    global.screen.connect('toggle-recording', function() {
+        if (recorder == null) {
+            recorder = new Shell.Recorder({ stage: global.stage });
+        }
+
+        if (recorder.is_recording()) {
+            recorder.pause();
+            Meta.enable_unredirect_for_screen(global.screen);
+        } else {
+            // read the parameters from GSettings always in case they have changed
+            recorder.set_framerate(recorderSettings.get_int('framerate'));
+            recorder.set_filename('shell-%d%u-%c.' + recorderSettings.get_string('file-extension'));
+            let pipeline = recorderSettings.get_string('pipeline');
+
+            if (!pipeline.match(/^\s*$/))
+                recorder.set_pipeline(pipeline);
+            else
+                recorder.set_pipeline(null);
+
+            Meta.disable_unredirect_for_screen(global.screen);
+            recorder.record();
+        }
+    });
+}
+
+function _initUserSession() {
+    _initRecorder();
+
+    keyboard.init();
+
+    global.screen.override_workspace_layout(Meta.ScreenCorner.TOPLEFT, false, -1, 1);
+
+    ExtensionSystem.init();
+    ExtensionSystem.loadExtensions();
+
+    let shellwm = global.window_manager;
+
+    shellwm.takeover_keybinding('panel_run_dialog');
+    shellwm.connect('keybinding::panel_run_dialog', function () {
+       getRunDialog().open();
+    });
+
+    shellwm.takeover_keybinding('panel_main_menu');
+    shellwm.connect('keybinding::panel_main_menu', function () {
+        overview.toggle();
+    });
+
+    global.display.connect('overlay-key', Lang.bind(overview, overview.toggle));
+
+}
+
 function start() {
     // Monkey patch utility functions into the global proxy;
     // This is easier and faster than indirecting down into global
@@ -85,10 +171,6 @@ function start() {
     // back into sync ones.
     DBus.session.flush();
 
-    // Load the calendar server. Note that we are careful about
-    // not loading any events until the user presses the clock
-    global.launch_calendar_server();
-
     // Ensure ShellWindowTracker and ShellAppUsage are initialized; this will
     // also initialize ShellAppSystem first.  ShellAppSystem
     // needs to load all the .desktop files, and ShellWindowTracker
@@ -107,18 +189,9 @@ function start() {
     global.stage.no_clear_hint = true;
 
     _defaultCssStylesheet = global.datadir + '/theme/gnome-shell.css';
+    _gdmCssStylesheet = global.datadir + '/theme/gdm.css';
     loadTheme();
 
-    let shellwm = global.window_manager;
-    shellwm.takeover_keybinding('panel_main_menu');
-    shellwm.connect('keybinding::panel_main_menu', function () {
-        overview.toggle();
-    });
-    shellwm.takeover_keybinding('panel_run_dialog');
-    shellwm.connect('keybinding::panel_run_dialog', function () {
-       getRunDialog().open();
-    });
-
     // Set up stage hierarchy to group all UI actors under one container.
     uiGroup = new Clutter.Group();
     St.set_ui_root(global.stage, uiGroup);
@@ -126,51 +199,34 @@ function start() {
     global.overlay_group.reparent(uiGroup);
     global.stage.add_actor(uiGroup);
 
-    placesManager = new PlaceDisplay.PlacesManager();
+    layoutManager = new Layout.LayoutManager();
     xdndHandler = new XdndHandler.XdndHandler();
     ctrlAltTabManager = new CtrlAltTab.CtrlAltTabManager();
-    overview = new Overview.Overview();
-    chrome = new Chrome.Chrome();
+    // This overview object is just a stub for non-user sessions
+    overview = new Overview.Overview({ isDummy: global.session_type != Shell.SessionType.USER });
     magnifier = new Magnifier.Magnifier();
     statusIconDispatcher = new StatusIconDispatcher.StatusIconDispatcher();
     panel = new Panel.Panel();
     wm = new WindowManager.WindowManager();
     messageTray = new MessageTray.MessageTray();
+    keyboard = new Keyboard.Keyboard();
     notificationDaemon = new NotificationDaemon.NotificationDaemon();
     windowAttentionHandler = new WindowAttentionHandler.WindowAttentionHandler();
-    telepathyClient = new TelepathyClient.Client();
 
+    if (global.session_type == Shell.SessionType.USER)
+        _createUserSession();
+    else if (global.session_type == Shell.SessionType.GDM)
+        _createGDMSession();
+
+    panel.startStatusArea();
+
+    keyboard.init();
     overview.init();
+
+    if (global.session_type == Shell.SessionType.USER)
+        _initUserSession();
     statusIconDispatcher.start(messageTray.actor);
 
-    _startDate = new Date();
-
-    let recorderSettings = new Gio.Settings({ schema: 'org.gnome.shell.recorder' });
-
-    global.screen.connect('toggle-recording', function() {
-        if (recorder == null) {
-            recorder = new Shell.Recorder({ stage: global.stage });
-        }
-
-        if (recorder.is_recording()) {
-            recorder.pause();
-        } else {
-            // read the parameters from GSettings always in case they have changed
-            recorder.set_framerate(recorderSettings.get_int('framerate'));
-            recorder.set_filename('shell-%d%u-%c.' + recorderSettings.get_string('file-extension'));
-            let pipeline = recorderSettings.get_string('pipeline');
-
-            if (!pipeline.match(/^\s*$/))
-                recorder.set_pipeline(pipeline);
-            else
-                recorder.set_pipeline(null);
-
-            recorder.record();
-        }
-    });
-
-    global.screen.override_workspace_layout(Meta.ScreenCorner.TOPLEFT, false, -1, 1);
-
     // Provide the bus object for gnome-session to
     // initiate logouts.
     EndSessionDialog.init();
@@ -178,19 +234,7 @@ function start() {
     // Attempt to become a PolicyKit authentication agent
     PolkitAuthenticationAgent.init()
 
-    global.screen.connect('monitors-changed', _relayout);
-
-    ExtensionSystem.init();
-    ExtensionSystem.loadExtensions();
-
-    // Perform initial relayout here
-    _relayout();
-
-    panel.startStatusArea();
-    panel.startupAnimation();
-
-    let display = global.screen.get_display();
-    display.connect('overlay-key', Lang.bind(overview, overview.toggle));
+    _startDate = new Date();
 
     global.stage.connect('captured-event', _globalKeyPressHandler);
 
@@ -300,14 +344,14 @@ function _windowRemoved(workspace, window) {
 function _windowLeftMonitor(metaScreen, monitorIndex, metaWin) {
     // If the window left the primary monitor, that
     // might make that workspace empty
-    if (monitorIndex == global.get_primary_monitor_index())
+    if (monitorIndex == layoutManager.primaryIndex)
         _queueCheckWorkspaces();
 }
 
 function _windowEnteredMonitor(metaScreen, monitorIndex, metaWin) {
     // If the window entered the primary monitor, that
     // might make that workspace non-empty
-    if (monitorIndex == global.get_primary_monitor_index())
+    if (monitorIndex == layoutManager.primaryIndex)
         _queueCheckWorkspaces();
 }
 
@@ -409,6 +453,9 @@ function loadTheme() {
 
     let theme = new St.Theme ({ application_stylesheet: cssStylesheet });
 
+    if (global.session_type == Shell.SessionType.GDM)
+        theme.load_stylesheet(_gdmCssStylesheet);
+
     if (previousTheme) {
         let customStylesheets = previousTheme.get_custom_stylesheets();
 
@@ -419,6 +466,19 @@ function loadTheme() {
     themeContext.set_theme (theme);
 }
 
+/**
+ * notify:
+ * @msg: A message
+ * @details: Additional information
+ */
+function notify(msg, details) {
+    let source = new MessageTray.SystemNotificationSource();
+    messageTray.add(source);
+    let notification = new MessageTray.Notification(source, msg, details);
+    notification.setTransient(true);
+    source.notify(notification);
+}
+
 /**
  * notifyError:
  * @msg: An error message
@@ -433,11 +493,7 @@ function notifyError(msg, details) {
     else
         log("error: " + msg)
 
-    let source = new MessageTray.SystemNotificationSource();
-    messageTray.add(source);
-    let notification = new MessageTray.Notification(source, msg, details);
-    notification.setTransient(true);
-    source.notify(notification);
+    notify(msg, details);
 }
 
 /**
@@ -481,80 +537,16 @@ function _getAndClearErrorStack() {
     return errors;
 }
 
-function _relayout() {
-    let monitors = global.get_monitors();
-    // destroy old corners
-    for (let i = 0; i < hotCorners.length; i++)
-        hotCorners[i].destroy();
-    hotCorners = [];
-
-
-    let primary = global.get_primary_monitor();
-    for (let i = 0; i < monitors.length; i++) {
-        let monitor = monitors[i];
-        let isPrimary = (monitor.x == primary.x &&
-                         monitor.y == primary.y &&
-                         monitor.width == primary.width &&
-                         monitor.height == primary.height);
-
-        let cornerX = monitor.x;
-        let cornerY = monitor.y;
-        if (St.Widget.get_default_direction() == St.TextDirection.RTL)
-            cornerX += monitor.width;
-
-
-        let haveTopLeftCorner = true;
-
-        /* Check if we have a top left (right for RTL) corner.
-         * I.e. if there is no monitor directly above or to the left(right) */
-        let besideX;
-        if (St.Widget.get_default_direction() == St.TextDirection.RTL)
-            besideX = monitor.x + 1;
-        else
-            besideX = cornerX - 1;
-        let besideY = cornerY;
-        let aboveX = cornerX;
-        let aboveY = cornerY - 1;
-
-        for (let j = 0; j < monitors.length; j++) {
-            if (i == j)
-                continue;
-            let otherMonitor = monitors[j];
-            if (besideX >= otherMonitor.x &&
-                besideX < otherMonitor.x + otherMonitor.width &&
-                besideY >= otherMonitor.y &&
-                besideY < otherMonitor.y + otherMonitor.height) {
-                haveTopLeftCorner = false;
-                break;
-            }
-            if (aboveX >= otherMonitor.x &&
-                aboveX < otherMonitor.x + otherMonitor.width &&
-                aboveY >= otherMonitor.y &&
-                aboveY < otherMonitor.y + otherMonitor.height) {
-                haveTopLeftCorner = false;
-                break;
-            }
-        }
-
-        /* We only want hot corners where there is a natural top-left
-         * corner, and on the primary monitor */
-        if (!isPrimary && !haveTopLeftCorner)
-            continue;
-
-        let corner = new Panel.HotCorner(isPrimary ? panel.button : null);
-        hotCorners.push(corner);
-        corner.actor.set_position(cornerX, cornerY);
-        if (isPrimary)
-            panel.setHotCorner(corner);
+function logStackTrace(msg) {
+    try {
+        throw new Error();
+    } catch (e) {
+        // e.stack must have at least two lines, with the first being
+        // logStackTrace() (which we strip off), and the second being
+        // our caller.
+        let trace = e.stack.substr(e.stack.indexOf('\n') + 1);
+        log(msg ? (msg + '\n' + trace) : trace);
     }
-
-    panel.relayout();
-    overview.relayout();
-
-    // To avoid updating the position and size of the workspaces
-    // in the overview, we just hide the overview. The positions
-    // will be updated when it is next shown.
-    overview.hide();
 }
 
 function isWindowActorDisplayedOnWorkspace(win, workspaceIndex) {
@@ -583,9 +575,8 @@ function _globalKeyPressHandler(actor, event) {
     let keyCode = event.get_key_code();
     let modifierState = Shell.get_event_state(event);
 
-    let display = global.screen.get_display();
     // This relies on the fact that Clutter.ModifierType is the same as Gdk.ModifierType
-    let action = display.get_keybinding_action(keyCode, modifierState);
+    let action = global.display.get_keybinding_action(keyCode, modifierState);
 
     // The screenshot action should always be available (even if a
     // modal dialog is present)
@@ -608,6 +599,15 @@ function _globalKeyPressHandler(actor, event) {
         return true;
     }
 
+    if (action == Meta.KeyBindingAction.SWITCH_PANELS) {
+        ctrlAltTabManager.popup(modifierState & Clutter.ModifierType.SHIFT_MASK);
+        return true;
+    }
+
+    // None of the other bindings are relevant outside of the user's session
+    if (global.session_type != Shell.SessionType.USER)
+        return false;
+
     switch (action) {
         // left/right would effectively act as synonyms for up/down if we enabled them;
         // but that could be considered confusing; we also disable them in the main view.
@@ -631,9 +631,6 @@ function _globalKeyPressHandler(actor, event) {
         case Meta.KeyBindingAction.PANEL_MAIN_MENU:
             overview.hide();
             return true;
-        case Meta.KeyBindingAction.SWITCH_PANELS:
-            ctrlAltTabManager.popup(modifierState & Clutter.ModifierType.SHIFT_MASK);
-            return true;
     }
 
     return false;
