@@ -20,9 +20,6 @@ const Params = imports.misc.params;
 const Tweener = imports.ui.tweener;
 const Util = imports.misc.util;
 
-const Gettext = imports.gettext.domain('gnome-shell');
-const _ = Gettext.gettext;
-
 const ANIMATION_TIME = 0.2;
 const NOTIFICATION_TIMEOUT = 4;
 const SUMMARY_TIMEOUT = 1;
@@ -71,14 +68,19 @@ function _fixMarkup(text, allowMarkup) {
         // Support &amp;, &quot;, &apos;, &lt; and &gt;, escape all other
         // occurrences of '&'.
         let _text = text.replace(/&(?!amp;|quot;|apos;|lt;|gt;)/g, '&amp;');
+
         // Support <b>, <i>, and <u>, escape anything else
         // so it displays as raw markup.
-        return _text.replace(/<(\/?[^biu]>|[^>\/][^>])/g, '&lt;$1');
-    } else {
-        // Escape everything
-        let _text = text.replace(/&/g, '&amp;');
-        return _text.replace(/</g, '&lt;');
+        _text = _text.replace(/<(?!\/?[biu]>)/g, '&lt;');
+
+        try {
+            Pango.parse_markup(_text, -1, '');
+            return _text;
+        } catch (e) {}
     }
+
+    // !allowMarkup, or invalid markup
+    return GLib.markup_escape_text(text, -1);
 }
 
 function URLHighlighter(text, lineWrap, allowMarkup) {
@@ -232,9 +234,7 @@ FocusGrabber.prototype = {
 
         this.actor = actor;
 
-        let metaDisplay = global.screen.get_display();
-
-        this._prevFocusedWindow = metaDisplay.focus_window;
+        this._prevFocusedWindow = global.display.focus_window;
         this._prevKeyFocusActor = global.stage.get_key_focus();
 
         if (!Main.overview.visible)
@@ -269,7 +269,7 @@ FocusGrabber.prototype = {
         let source = event.get_source();
         switch (event.type()) {
             case Clutter.EventType.BUTTON_PRESS:
-                if (!this.actor.contains(source))
+                if (!this.actor.contains(source) && !Main.keyboard.actor.contains(source))
                     this.emit('button-pressed', source);
                 break;
             case Clutter.EventType.KEY_PRESS:
@@ -288,8 +288,6 @@ FocusGrabber.prototype = {
         if (!this._hasFocus)
             return;
 
-        let metaDisplay = global.screen.get_display();
-
         if (this._focusActorChangedId > 0) {
             global.stage.disconnect(this._focusActorChangedId);
             this._focusActorChangedId = 0;
@@ -308,8 +306,8 @@ FocusGrabber.prototype = {
         this._hasFocus = false;
         this.emit('focus-ungrabbed');
 
-        if (this._prevFocusedWindow && !metaDisplay.focus_window) {
-            metaDisplay.set_input_focus_window(this._prevFocusedWindow, false, global.get_current_time());
+        if (this._prevFocusedWindow && !global.display.focus_window) {
+            global.display.set_input_focus_window(this._prevFocusedWindow, false, global.get_current_time());
             this._prevFocusedWindow = null;
         }
         if (this._prevKeyFocusActor) {
@@ -398,6 +396,8 @@ function Notification(source, title, banner, params) {
 }
 
 Notification.prototype = {
+    IMAGE_SIZE: 125,
+
     _init: function(source, title, banner, params) {
         this.source = source;
         this.urgency = Urgency.NORMAL;
@@ -414,6 +414,7 @@ Notification.prototype = {
         this._titleDirection = St.TextDirection.NONE;
         this._spacing = 0;
         this._scrollPolicy = Gtk.PolicyType.AUTOMATIC;
+        this._imageBin = null;
 
         source.connect('destroy', Lang.bind(this,
             function (source, reason) {
@@ -443,9 +444,19 @@ Notification.prototype = {
         this._bannerBox.connect('allocate', Lang.bind(this, this._bannerBoxAllocate));
         this._table.add(this._bannerBox, { row: 0,
                                            col: 1,
+                                           col_span: 2,
+                                           x_expand: false,
                                            y_expand: false,
                                            y_fill: false });
 
+        // This is an empty cell that overlaps with this._bannerBox cell to ensure
+        // that this._bannerBox cell expands horizontally, while not forcing the
+        // this._imageBin that is also in col: 2 to expand horizontally.
+        this._table.add(new St.Bin(), { row: 0,
+                                        col: 2,
+                                        y_expand: false,
+                                        y_fill: false });
+
         this._titleLabel = new St.Label();
         this._bannerBox.add_actor(this._titleLabel);
         this._bannerUrlHighlighter = new URLHighlighter();
@@ -497,7 +508,10 @@ Notification.prototype = {
             this._actionArea = null;
             this._buttonBox = null;
         }
-        if (!this._scrollArea && !this._actionArea)
+        if (this._imageBin && params.clear)
+            this.unsetImage();
+
+        if (!this._scrollArea && !this._actionArea && !this._imageBin)
             this._table.remove_style_class_name('multi-line-notification');
 
         this._icon = params.icon || this.source.createNotificationIcon();
@@ -560,8 +574,10 @@ Notification.prototype = {
         this._scrollArea = new St.ScrollView({ name: 'notification-scrollview',
                                                vscrollbar_policy: this._scrollPolicy,
                                                hscrollbar_policy: Gtk.PolicyType.NEVER,
-                                               vfade: true });
-        this._table.add(this._scrollArea, { row: 1, col: 1 });
+                                               style_class: 'vfade' });
+        this._table.add(this._scrollArea, { row: 1,
+                                            col: 2 });
+        this._updateLastColumnSettings();
         this._contentArea = new St.BoxLayout({ name: 'notification-body',
                                                vertical: true });
         this._scrollArea.add_actor(this._contentArea);
@@ -638,13 +654,52 @@ Notification.prototype = {
         if (!props)
             props = {};
         props.row = 2;
-        props.col = 1;
+        props.col = 2;
 
         this._table.add_style_class_name('multi-line-notification');
         this._table.add(this._actionArea, props);
+        this._updateLastColumnSettings();
         this._updated();
     },
 
+    _updateLastColumnSettings: function() {
+        if (this._scrollArea)
+            this._table.child_set(this._scrollArea, { col: this._imageBin ? 2 : 1,
+                                                      col_span: this._imageBin ? 1 : 2 });
+        if (this._actionArea)
+            this._table.child_set(this._actionArea, { col: this._imageBin ? 2 : 1,
+                                                      col_span: this._imageBin ? 1 : 2 });
+    },
+
+    setImage: function(image) {
+        if (this._imageBin)
+            this.unsetImage();
+        this._imageBin = new St.Bin();
+        this._imageBin.child = image;
+        this._imageBin.opacity = 230;
+        this._table.add_style_class_name('multi-line-notification');
+        this._table.add_style_class_name('notification-with-image');
+        this._updateLastColumnSettings();
+        this._table.add(this._imageBin, { row: 1,
+                                          col: 1,
+                                          row_span: 2,
+                                          x_expand: false,
+                                          y_expand: false,
+                                          x_fill: false,
+                                          y_fill: false });
+    },
+
+    unsetImage: function() {
+        if (this._imageBin) {
+            this._table.remove_style_class_name('notification-with-image');
+            this._table.remove_actor(this._imageBin);
+            this._imageBin = null;
+            this._updateLastColumnSettings();
+            if (!this._scrollArea && !this._actionArea)
+                this._table.remove_style_class_name('multi-line-notification');
+        }
+    },
+
     // addButton:
     // @id: the action ID
     // @label: the label for the action's button
@@ -660,7 +715,9 @@ Notification.prototype = {
 
             let box = new St.BoxLayout({ name: 'notification-actions' });
             this.setActionArea(box, { x_expand: false,
+                                      y_expand: false,
                                       x_fill: false,
+                                      y_fill: false,
                                       x_align: St.Align.END });
             this._buttonBox = box;
         }
@@ -832,6 +889,7 @@ Notification.prototype = {
         // Restore banner opacity in case the notification is shown in the
         // banner mode again on update.
         this._bannerLabel.opacity = 255;
+        this.emit('collapsed');
     },
 
     _onActionInvoked: function(actor, mouseButtonClicked, id) {
@@ -882,20 +940,96 @@ Source.prototype = {
 
     _init: function(title) {
         this.title = title;
+
+        this.actor = new Shell.GenericContainer();
+        this.actor.connect('get-preferred-width', Lang.bind(this, this._getPreferredWidth));
+        this.actor.connect('get-preferred-height', Lang.bind(this, this._getPreferredHeight));
+        this.actor.connect('allocate', Lang.bind(this, this._allocate));
+        this.actor.connect('destroy', Lang.bind(this,
+            function() {
+                this._actorDestroyed = true;
+            }));
+        this._actorDestroyed = false;
+
+        this._counterLabel = new St.Label();
+        this._counterBin = new St.Bin({ style_class: 'summary-source-counter',
+                                        child: this._counterLabel });
+        this._counterBin.hide();
+
         this._iconBin = new St.Bin({ width: this.ICON_SIZE,
                                      height: this.ICON_SIZE,
                                      x_fill: true,
                                      y_fill: true });
+
+        this.actor.add_actor(this._iconBin);
+        this.actor.add_actor(this._counterBin);
+
         this.isTransient = false;
         this.isChat = false;
 
         this.notifications = [];
     },
 
+    _getPreferredWidth: function (actor, forHeight, alloc) {
+        let [min, nat] = this._iconBin.get_preferred_width(forHeight);
+        alloc.min_size = min; alloc.nat_size = nat;
+    },
+
+    _getPreferredHeight: function (actor, forWidth, alloc) {
+        let [min, nat] = this._iconBin.get_preferred_height(forWidth);
+        alloc.min_size = min; alloc.nat_size = nat;
+    },
+
+    _allocate: function(actor, box, flags) {
+        // the iconBin should fill our entire box
+        this._iconBin.allocate(box, flags);
+
+        let childBox = new Clutter.ActorBox();
+
+        let [minWidth, minHeight, naturalWidth, naturalHeight] = this._counterBin.get_preferred_size();
+        let direction = this.actor.get_direction();
+
+        if (direction == St.TextDirection.LTR) {
+            // allocate on the right in LTR
+            childBox.x1 = box.x2 - naturalWidth;
+            childBox.x2 = box.x2;
+        } else {
+            // allocate on the left in RTL
+            childBox.x1 = 0;
+            childBox.x2 = naturalWidth;
+        }
+
+        childBox.y1 = box.y2 - naturalHeight;
+        childBox.y2 = box.y2;
+
+        this._counterBin.allocate(childBox, flags);
+    },
+
+    _setCount: function(count, visible) {
+        if (isNaN(parseInt(count)))
+            throw new Error("Invalid notification count: " + count);
+
+        if (this._actorDestroyed)
+            return;
+
+        this._counterBin.visible = visible;
+        this._counterLabel.set_text(count.toString());
+    },
+
+    _updateCount: function() {
+        let count = this.notifications.length;
+        this._setCount(count, count > 1);
+    },
+
     setTransient: function(isTransient) {
         this.isTransient = isTransient;
     },
 
+    setTitle: function(newTitle) {
+        this.title = newTitle;
+        this.emit('title-changed');
+    },
+
     // Called to create a new icon actor (of size this.ICON_SIZE).
     // Must be overridden by the subclass if you do not pass icons
     // explicitly to the Notification() constructor.
@@ -906,7 +1040,7 @@ Source.prototype = {
     // Unlike createNotificationIcon, this always returns the same actor;
     // there is only one summary icon actor for a Source.
     getSummaryIcon: function() {
-        return this._iconBin;
+        return this.actor;
     },
 
     pushNotification: function(notification) {
@@ -925,7 +1059,11 @@ Source.prototype = {
                 this.notifications.splice(index, 1);
                 if (this.notifications.length == 0)
                     this._lastNotificationRemoved();
+
+                this._updateCount();
             }));
+
+        this._updateCount();
     },
 
     notify: function(notification) {
@@ -962,6 +1100,8 @@ Source.prototype = {
         for (let i = this.notifications.length - 1; i >= 0; i--)
             if (!this.notifications[i].resident)
                 this.notifications[i].destroy();
+
+        this._updateCount();
     },
 
     // Default implementation is to destroy this source, but subclasses can override
@@ -998,6 +1138,11 @@ SummaryItem.prototype = {
         this._sourceTitleBin.child = this._sourceTitle;
         this._sourceTitleBin.width = 0;
 
+        this.source.connect('title-changed',
+                            Lang.bind(this, function() {
+                                this._sourceTitle.text = source.title;
+                            }));
+
         this._sourceBox.add(this._sourceIcon, { y_fill: false });
         this._sourceBox.add(this._sourceTitleBin, { expand: true, y_fill: false });
         this.actor.child = this._sourceBox;
@@ -1005,7 +1150,7 @@ SummaryItem.prototype = {
         this.notificationStackView = new St.ScrollView({ name: source.isChat ? '' : 'summary-notification-stack-scrollview',
                                                          vscrollbar_policy: source.isChat ? Gtk.PolicyType.NEVER : Gtk.PolicyType.AUTOMATIC,
                                                          hscrollbar_policy: Gtk.PolicyType.NEVER,
-                                                         vfade: true });
+                                                         style_class: 'vfade' });
         this.notificationStack = new St.BoxLayout({ name: 'summary-notification-stack',
                                                      vertical: true });
         this.notificationStackView.add_actor(this.notificationStack);
@@ -1189,12 +1334,11 @@ MessageTray.prototype = {
         this._summaryMotionId = 0;
 
         this._summaryBoxPointer = new BoxPointer.BoxPointer(St.Side.BOTTOM,
-                                                           { reactive: true,
-                                                             track_hover: true });
+                                                            { reactive: true,
+                                                              track_hover: true });
         this._summaryBoxPointer.actor.style_class = 'summary-boxpointer';
-        this.actor.add_actor(this._summaryBoxPointer.actor);
-        this._summaryBoxPointer.actor.lower_bottom();
         this._summaryBoxPointer.actor.hide();
+        Main.layoutManager.addChrome(this._summaryBoxPointer.actor, { visibleInFullscreen: true });
 
         this._summaryBoxPointerItem = null;
         this._summaryBoxPointerContentUpdatedId = 0;
@@ -1229,6 +1373,7 @@ MessageTray.prototype = {
 
         this._trayState = State.HIDDEN;
         this._locked = false;
+        this._traySummoned = false;
         this._useLongerTrayLeftTimeout = false;
         this._trayLeftTimeoutId = 0;
         this._pointerInTray = false;
@@ -1244,12 +1389,12 @@ MessageTray.prototype = {
         this._notificationRemoved = false;
         this._reNotifyAfterHideNotification = null;
 
-        Main.chrome.addActor(this.actor, { affectsStruts: false,
-                                           visibleInOverview: true });
-        Main.chrome.trackActor(this._notificationBin);
-        Main.chrome.trackActor(this._summaryBoxPointer.actor);
+        Main.layoutManager.trayBox.add_actor(this.actor);
+        this.actor.y = -1;
+        Main.layoutManager.trackChrome(this.actor);
+        Main.layoutManager.trackChrome(this._notificationBin);
 
-        global.screen.connect('monitors-changed', Lang.bind(this, this._setSizePosition));
+        Main.layoutManager.connect('monitors-changed', Lang.bind(this, this._setSizePosition));
 
         this._setSizePosition();
 
@@ -1285,23 +1430,11 @@ MessageTray.prototype = {
     },
 
     _setSizePosition: function() {
-        let primary = global.get_primary_monitor();
-        this.actor.x = primary.x;
-        this.actor.y = primary.y + primary.height - 1;
-        this.actor.width = primary.width;
+        let monitor = Main.layoutManager.bottomMonitor;
         this._notificationBin.x = 0;
-        this._notificationBin.width = primary.width;
+        this._notificationBin.width = monitor.width;
         this._summaryBin.x = 0;
-        this._summaryBin.width = primary.width;
-
-        if (this._pointerBarrier)
-            global.destroy_pointer_barrier(this._pointerBarrier);
-        this._pointerBarrier =
-            global.create_pointer_barrier(primary.x + primary.width, primary.y + primary.height - this.actor.height,
-                                          primary.x + primary.width, primary.y + primary.height,
-                                          4 /* BarrierNegativeX */);
-
-
+        this._summaryBin.width = monitor.width;
     },
 
     contains: function(source) {
@@ -1447,7 +1580,19 @@ MessageTray.prototype = {
         if (!this._locked)
             return;
         this._locked = false;
-        this._pointerInTray = this.actor.hover && !this._summaryBoxPointer.bin.hover;
+        this._pointerInTray = this.actor.hover;
+        this._updateState();
+    },
+
+    toggle: function() {
+        this._traySummoned = !this._traySummoned;
+        this._updateState();
+    },
+
+    hide: function() {
+        this._traySummoned = false;
+        this.actor.set_hover(false);
+        this._summary.set_hover(false);
         this._updateState();
     },
 
@@ -1608,6 +1753,8 @@ MessageTray.prototype = {
                  this._clickedSummaryItemMouseButton != button) {
             this._clickedSummaryItem = summaryItem;
             this._clickedSummaryItemMouseButton = button;
+
+            summaryItem.source.emit('summary-item-clicked', button);
         } else {
             this._unsetClickedSummaryItem();
         }
@@ -1633,13 +1780,6 @@ MessageTray.prototype = {
             if (this._useLongerTrayLeftTimeout && !this._trayLeftTimeoutId)
                 return;
 
-            // Don't do anything if the mouse is over the summary notification as this should be considered as
-            // leaving the tray. The tray is locked when the summary notification is visible anyway, but we
-            // should treat the mouse being over the summary notification as the tray being left for collapsing
-            // any expanded summary item other than the one related to the notification.
-            if (this._summaryBoxPointer.bin.hover)
-                return;
-
             this._useLongerTrayLeftTimeout = false;
             if (this._trayLeftTimeoutId) {
                 Mainloop.source_remove(this._trayLeftTimeoutId);
@@ -1763,7 +1903,7 @@ MessageTray.prototype = {
         }
 
         // Summary
-        let summarySummoned = this._pointerInSummary || this._overviewVisible;
+        let summarySummoned = this._pointerInSummary || this._overviewVisible ||  this._traySummoned;
         let summaryPinned = this._summaryTimeoutId != 0 || this._pointerInTray || summarySummoned || this._locked;
         let summaryHovered = this._pointerInTray || this._pointerInSummary;
         let summaryVisibleWithNoHover = (this._overviewVisible || this._locked) && !summaryHovered;
@@ -1857,18 +1997,16 @@ MessageTray.prototype = {
     },
 
     _showTray: function() {
-        let primary = global.get_primary_monitor();
         this._tween(this.actor, '_trayState', State.SHOWN,
-                    { y: primary.y + primary.height - this.actor.height,
+                    { y: -this.actor.height,
                       time: ANIMATION_TIME,
                       transition: 'easeOutQuad'
                     });
     },
 
     _hideTray: function() {
-        let primary = global.get_primary_monitor();
         this._tween(this.actor, '_trayState', State.HIDDEN,
-                    { y: primary.y + primary.height - 1,
+                    { y: -1,
                       time: ANIMATION_TIME,
                       transition: 'easeOutQuad'
                     });
@@ -1956,7 +2094,7 @@ MessageTray.prototype = {
 
     _notificationTimeout: function() {
         let [x, y, mods] = global.get_pointer();
-        if (y > this._lastSeenMouseY + 10 && y < this.actor.y) {
+        if (y > this._lastSeenMouseY + 10 && !this.actor.hover) {
             // The mouse is moving towards the notification, so don't
             // hide it yet. (We just create a new timeout (and destroy
             // the old one) each time because the bookkeeping is
@@ -2014,9 +2152,15 @@ MessageTray.prototype = {
         this._notification.expand(!autoExpanding);
     },
 
-   _onNotificationExpanded: function() {
+    _onNotificationExpanded: function() {
         let expandedY = this.actor.height - this._notificationBin.height;
-        if (this._notificationBin.y != expandedY)
+
+        // Don't animate the notification to its new position if it has shrunk:
+        // there will be a very visible "gap" that breaks the illusion.
+
+        if (this._notificationBin.y < expandedY)
+            this._notificationBin.y = expandedY;
+        else if (this._notification.y != expandedY)
             this._tween(this._notificationBin, '_notificationState', State.SHOWN,
                         { y: expandedY,
                           time: ANIMATION_TIME,
@@ -2031,7 +2175,6 @@ MessageTray.prototype = {
     },
 
     _showSummary: function(timeout) {
-        let primary = global.get_primary_monitor();
         this._summaryBin.opacity = 0;
         this._summaryBin.y = this.actor.height;
         this._tween(this._summaryBin, '_summaryState', State.SHOWN,
