@@ -11,6 +11,7 @@ const Shell = imports.gi.Shell;
 const Gdk = imports.gi.Gdk;
 
 const AppDisplay = imports.ui.appDisplay;
+const ContactDisplay = imports.ui.contactDisplay;
 const Dash = imports.ui.dash;
 const DND = imports.ui.dnd;
 const DocDisplay = imports.ui.docDisplay;
@@ -18,6 +19,7 @@ const Lightbox = imports.ui.lightbox;
 const Main = imports.ui.main;
 const MessageTray = imports.ui.messageTray;
 const Panel = imports.ui.panel;
+const Params = imports.misc.params;
 const PlaceDisplay = imports.ui.placeDisplay;
 const Tweener = imports.ui.tweener;
 const ViewSelector = imports.ui.viewSelector;
@@ -96,14 +98,30 @@ ShellInfo.prototype = {
 };
 
 function Overview() {
-    this._init();
+    this._init.apply(this, arguments);
 }
 
 Overview.prototype = {
-    _init : function() {
-        // The actual global.background_actor is inside global.window_group,
-        // which is hidden when displaying the overview, so we display a clone.
-        this._background = new Clutter.Clone({ source: global.background_actor });
+    _init : function(params) {
+        params = Params.parse(params, { isDummy: false });
+
+        this.isDummy = params.isDummy;
+
+        // We only have an overview in user sessions, so
+        // create a dummy overview in other cases
+        if (this.isDummy) {
+            this.animationInProgress = false;
+            this.visible = false;
+            this.workspaces = null;
+            return;
+        }
+
+        // The main BackgroundActor is inside global.window_group which is
+        // hidden when displaying the overview, so we create a new
+        // one. Instances of this class share a single CoglTexture behind the
+        // scenes which allows us to show the background with different
+        // rendering options without duplicating the texture data.
+        this._background = Meta.BackgroundActor.new_for_screen(global.screen);
         this._background.hide();
         global.overlay_group.add_actor(this._background);
 
@@ -175,37 +193,53 @@ Overview.prototype = {
     // signal handlers and so forth. So we create them after
     // construction in this init() method.
     init: function() {
-        this.shellInfo = new ShellInfo();
+        if (this.isDummy)
+            return;
 
-        this.viewSelector = new ViewSelector.ViewSelector();
-        this._group.add_actor(this.viewSelector.actor);
+        this._shellInfo = new ShellInfo();
+
+        this._viewSelector = new ViewSelector.ViewSelector();
+        this._group.add_actor(this._viewSelector.actor);
 
         this._workspacesDisplay = new WorkspacesView.WorkspacesDisplay();
-        this.viewSelector.addViewTab('windows', _("Windows"), this._workspacesDisplay.actor, 'text-x-generic');
+        this._viewSelector.addViewTab('windows', _("Windows"), this._workspacesDisplay.actor, 'text-x-generic');
 
         let appView = new AppDisplay.AllAppDisplay();
-        this.viewSelector.addViewTab('applications', _("Applications"), appView.actor, 'system-run');
+        this._viewSelector.addViewTab('applications', _("Applications"), appView.actor, 'system-run');
 
         // Default search providers
-        this.viewSelector.addSearchProvider(new AppDisplay.AppSearchProvider());
-        this.viewSelector.addSearchProvider(new AppDisplay.PrefsSearchProvider());
-        this.viewSelector.addSearchProvider(new PlaceDisplay.PlaceSearchProvider());
-        this.viewSelector.addSearchProvider(new DocDisplay.DocSearchProvider());
+        this._viewSelector.addSearchProvider(new AppDisplay.AppSearchProvider());
+        this._viewSelector.addSearchProvider(new AppDisplay.SettingsSearchProvider());
+        this._viewSelector.addSearchProvider(new PlaceDisplay.PlaceSearchProvider());
+        this._viewSelector.addSearchProvider(new DocDisplay.DocSearchProvider());
+        this._viewSelector.addSearchProvider(new ContactDisplay.ContactSearchProvider());
 
         // TODO - recalculate everything when desktop size changes
-        this.dash = new Dash.Dash();
-        this._group.add_actor(this.dash.actor);
-        this.dash.actor.add_constraint(this.viewSelector.constrainY);
-        this.dash.actor.add_constraint(this.viewSelector.constrainHeight);
+        this._dash = new Dash.Dash();
+        this._group.add_actor(this._dash.actor);
+        this._dash.actor.add_constraint(this._viewSelector.constrainY);
+        this._dash.actor.add_constraint(this._viewSelector.constrainHeight);
+        this.dashIconSize = this._dash.iconSize;
+        this._dash.connect('icon-size-changed',
+                           Lang.bind(this, function() {
+                               this.dashIconSize = this._dash.iconSize;
+                           }));
 
         // Translators: this is the name of the dock/favorites area on
         // the left of the overview
-        Main.ctrlAltTabManager.addGroup(this.dash.actor, _("Dash"), 'user-bookmarks');
+        Main.ctrlAltTabManager.addGroup(this._dash.actor, _("Dash"), 'user-bookmarks');
 
         Main.layoutManager.connect('monitors-changed', Lang.bind(this, this._relayout));
         this._relayout();
     },
 
+    setMessage: function(text, undoCallback, undoLabel) {
+        if (this.isDummy)
+            return;
+
+        this._shellInfo.setMessage(text, undoCallback, undoLabel);
+    },
+
     _onDragBegin: function() {
         DND.addDragMonitor(this._dragMonitor);
         // Remember the workspace we started from
@@ -275,6 +309,9 @@ Overview.prototype = {
     },
 
     setScrollAdjustment: function(adjustment, direction) {
+        if (this.isDummy)
+            return;
+
         this._scrollAdjustment = adjustment;
         if (this._scrollAdjustment == null)
             this._scrollDirection = SwipeScrollDirection.NONE;
@@ -467,15 +504,15 @@ Overview.prototype = {
         // Set the dash's x position - y is handled by a constraint
         let dashX;
         if (rtl) {
-            this.dash.actor.set_anchor_point_from_gravity(Clutter.Gravity.NORTH_EAST);
+            this._dash.actor.set_anchor_point_from_gravity(Clutter.Gravity.NORTH_EAST);
             dashX = primary.width;
         } else {
             dashX = 0;
         }
-        this.dash.actor.set_x(dashX);
+        this._dash.actor.set_x(dashX);
 
-        this.viewSelector.actor.set_position(viewX, viewY);
-        this.viewSelector.actor.set_size(viewWidth, viewHeight);
+        this._viewSelector.actor.set_position(viewX, viewY);
+        this._viewSelector.actor.set_size(viewWidth, viewHeight);
     },
 
     //// Public methods ////
@@ -508,6 +545,8 @@ Overview.prototype = {
     //
     // Animates the overview visible and grabs mouse and keyboard input
     show : function() {
+        if (this.isDummy)
+            return;
         if (this._shown)
             return;
         // Do this manually instead of using _syncInputMode, to handle failure
@@ -535,6 +574,9 @@ Overview.prototype = {
         //
         // If we switched to displaying the actors in the Overview rather than
         // clones of them, this would obviously no longer be necessary.
+        //
+        // Disable unredirection while in the overview
+        Meta.disable_unredirect_for_screen(global.screen);
         global.window_group.hide();
         this._group.show();
         this._background.show();
@@ -566,6 +608,12 @@ Overview.prototype = {
                            onCompleteScope: this
                          });
 
+        Tweener.addTween(this._background,
+                         { dim_factor: 0.4,
+                           time: ANIMATION_TIME,
+                           transition: 'easeOutQuad'
+                         });
+
         this._coverPane.raise_top();
         this._coverPane.show();
         this.emit('showing');
@@ -578,6 +626,9 @@ Overview.prototype = {
     // will result in the overview not being hidden until hideTemporarily() is
     // called.
     showTemporarily: function() {
+        if (this.isDummy)
+            return;
+
         if (this._shownTemporarily)
             return;
 
@@ -590,6 +641,9 @@ Overview.prototype = {
     //
     // Reverses the effect of show()
     hide: function() {
+        if (this.isDummy)
+            return;
+
         if (!this._shown)
             return;
 
@@ -608,6 +662,9 @@ Overview.prototype = {
     //
     // Reverses the effect of showTemporarily()
     hideTemporarily: function() {
+        if (this.isDummy)
+            return;
+
         if (!this._shownTemporarily)
             return;
 
@@ -619,6 +676,9 @@ Overview.prototype = {
     },
 
     toggle: function() {
+        if (this.isDummy)
+            return;
+
         if (this._shown)
             this.hide();
         else
@@ -684,6 +744,12 @@ Overview.prototype = {
                            onCompleteScope: this
                          });
 
+        Tweener.addTween(this._background,
+                         { dim_factor: 1.0,
+                           time: ANIMATION_TIME,
+                           transition: 'easeOutQuad'
+                         });
+
         this._coverPane.raise_top();
         this._coverPane.show();
         this.emit('hiding');
@@ -704,6 +770,9 @@ Overview.prototype = {
     },
 
     _hideDone: function() {
+        // Re-enable unredirection
+        Meta.enable_unredirect_for_screen(global.screen);
+
         global.window_group.show();
 
         this.workspaces.destroy();
