@@ -15,6 +15,17 @@ const Tweener = imports.ui.tweener;
 
 const SLIDER_SCROLL_STEP = 0.05; /* Slider scrolling step in % */
 
+function _ensureStyle(actor) {
+    if (actor.get_children) {
+        let children = actor.get_children();
+        for (let i = 0; i < children.length; i++)
+            _ensureStyle(children[i]);
+    }
+
+    if (actor instanceof St.Widget)
+        actor.ensure_style();
+}
+
 function PopupBaseMenuItem(params) {
     this._init(params);
 }
@@ -706,7 +717,8 @@ PopupSwitchMenuItem.prototype = {
         this.addActor(this.label);
 
         this._statusBin = new St.Bin({ x_align: St.Align.END });
-        this.addActor(this._statusBin, { align: St.Align.END });
+        this.addActor(this._statusBin,
+                      { expand: true, span: -1, align: St.Align.END });
 
         this._statusLabel = new St.Label({ text: '',
                                            style_class: 'popup-inactive-menu-item'
@@ -800,14 +812,58 @@ PopupMenuBase.prototype = {
         this.passEvents = false;
 
         this._activeMenuItem = null;
+        this._childMenus = [];
     },
 
     addAction: function(title, callback) {
-        var menuItem = new PopupMenuItem(title);
+        let menuItem = new PopupMenuItem(title);
         this.addMenuItem(menuItem);
         menuItem.connect('activate', Lang.bind(this, function (menuItem, event) {
             callback(event);
         }));
+
+        return menuItem;
+    },
+
+    addSettingsAction: function(title, desktopFile) {
+        // Don't allow user settings to get edited unless we're in a user session
+        if (global.session_type != Shell.SessionType.USER)
+            return null;
+
+        let menuItem = this.addAction(title, function() {
+                           let app = Shell.AppSystem.get_default().lookup_setting(desktopFile);
+
+                           if (!app) {
+                               log('Settings panel for desktop file ' + desktopFile + ' could not be loaded!');
+                               return;
+                           }
+
+                           Main.overview.hide();
+                           app.activate();
+                       });
+        return menuItem;
+    },
+
+    isChildMenu: function(menu) {
+        return this._childMenus.indexOf(menu) != -1;
+    },
+
+    addChildMenu: function(menu) {
+        if (this.isChildMenu(menu))
+            return;
+
+        this._childMenus.push(menu);
+        this.emit('child-menu-added', menu);
+    },
+
+    removeChildMenu: function(menu) {
+        let index = this._childMenus.indexOf(menu);
+
+        if (index == -1)
+            return;
+
+        this._childMenus.splice(index, 1);
+        this.emit('child-menu-removed', menu);
     },
 
     /**
@@ -860,6 +916,39 @@ PopupMenuBase.prototype = {
         }));
     },
 
+    _updateSeparatorVisibility: function(menuItem) {
+        let children = this.box.get_children();
+
+        let index = children.indexOf(menuItem.actor);
+
+        if (index < 0)
+            return;
+
+        let childBeforeIndex = index - 1;
+
+        while (childBeforeIndex >= 0 && !children[childBeforeIndex].visible)
+            childBeforeIndex--;
+
+        if (childBeforeIndex < 0
+            || children[childBeforeIndex]._delegate instanceof PopupSeparatorMenuItem) {
+            menuItem.actor.hide();
+            return;
+        }
+
+        let childAfterIndex = index + 1;
+
+        while (childAfterIndex < children.length && !children[childAfterIndex].visible)
+            childAfterIndex++;
+
+        if (childAfterIndex >= children.length
+            || children[childAfterIndex]._delegate instanceof PopupSeparatorMenuItem) {
+            menuItem.actor.hide();
+            return;
+        }
+
+        menuItem.actor.show();
+    },
+
     addMenuItem: function(menuItem, position) {
         let before_item = null;
         if (position == undefined) {
@@ -891,6 +980,14 @@ PopupMenuBase.prototype = {
                 if (!open)
                     menuItem.menu.close(false);
             });
+        } else if (menuItem instanceof PopupSeparatorMenuItem) {
+            this._connectItemSignals(menuItem);
+
+            // updateSeparatorVisibility needs to get called any time the
+            // separator's adjacent siblings change visibility or position.
+            // open-state-changed isn't exactly that, but doing it in more
+            // precise ways would require a lot more bookkeeping.
+            this.connect('open-state-changed', Lang.bind(this, function() { this._updateSeparatorVisibility(menuItem); }));
         } else if (menuItem instanceof PopupBaseMenuItem)
             this._connectItemSignals(menuItem);
         else
@@ -954,6 +1051,10 @@ PopupMenuBase.prototype = {
             return null;
     },
 
+    get numMenuItems() {
+        return this._getMenuItems().length;
+    },
+
     removeAll: function() {
         let children = this._getMenuItems();
         for (let i = 0; i < children.length; i++) {
@@ -985,12 +1086,11 @@ function PopupMenu() {
 PopupMenu.prototype = {
     __proto__: PopupMenuBase.prototype,
 
-    _init: function(sourceActor, alignment, arrowSide, gap) {
+    _init: function(sourceActor, alignment, arrowSide) {
         PopupMenuBase.prototype._init.call (this, sourceActor, 'popup-menu-content');
 
         this._alignment = alignment;
         this._arrowSide = arrowSide;
-        this._gap = gap;
 
         this._boxPointer = new BoxPointer.BoxPointer(arrowSide,
                                                      { x_fill: true,
@@ -1048,9 +1148,11 @@ PopupMenu.prototype = {
 
         this.isOpen = true;
 
-        this._boxPointer.setPosition(this.sourceActor, this._gap, this._alignment);
+        this._boxPointer.setPosition(this.sourceActor, this._alignment);
         this._boxPointer.show(animate);
 
+        this.actor.raise_top();
+
         this.emit('open-state-changed', true);
     },
 
@@ -1318,6 +1420,193 @@ PopupSubMenuMenuItem.prototype = {
     }
 };
 
+function PopupComboMenu() {
+    this._init.apply(this, arguments);
+}
+
+PopupComboMenu.prototype = {
+    __proto__: PopupMenuBase.prototype,
+
+    _init: function(sourceActor) {
+        PopupMenuBase.prototype._init.call(this,
+                                           sourceActor, 'popup-combo-menu');
+        this.actor = this.box;
+        this.actor._delegate = this;
+        this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
+        this.actor.connect('key-focus-in', Lang.bind(this, this._onKeyFocusIn));
+        this._activeItemPos = -1;
+        global.focus_manager.add_group(this.actor);
+    },
+
+    _onKeyPressEvent: function(actor, event) {
+        if (event.get_key_symbol() == Clutter.Escape) {
+            this.close(true);
+            return true;
+        }
+
+        return false;
+    },
+
+    _onKeyFocusIn: function(actor) {
+        let items = this._getMenuItems();
+        let activeItem = items[this._activeItemPos];
+        activeItem.actor.grab_key_focus();
+    },
+
+    open: function() {
+        if (this.isOpen)
+            return;
+
+        this.isOpen = true;
+
+        let [sourceX, sourceY] = this.sourceActor.get_transformed_position();
+        let items = this._getMenuItems();
+        let activeItem = items[this._activeItemPos];
+
+        this.actor.set_position(sourceX, sourceY - activeItem.actor.y);
+        this.actor.width = Math.max(this.actor.width, this.sourceActor.width);
+        this.actor.raise_top();
+
+        this.actor.opacity = 0;
+        this.actor.show();
+
+        Tweener.addTween(this.actor,
+                         { opacity: 255,
+                           transition: 'linear',
+                           time: BoxPointer.POPUP_ANIMATION_TIME });
+
+        this.emit('open-state-changed', true);
+    },
+
+    close: function() {
+        if (!this.isOpen)
+            return;
+
+        this.isOpen = false;
+        Tweener.addTween(this.actor,
+                         { opacity: 0,
+                           transition: 'linear',
+                           time: BoxPointer.POPUP_ANIMATION_TIME,
+                           onComplete: Lang.bind(this,
+                               function() {
+                                   this.actor.hide();
+                               })
+                         });
+
+        this.emit('open-state-changed', false);
+    },
+
+    setActiveItem: function(position) {
+        this._activeItemPos = position;
+    },
+
+    setItemVisible: function(position, visible) {
+        if (!visible && position == this._activeItemPos) {
+            log('Trying to hide the active menu item.');
+            return;
+        }
+
+        this._getMenuItems()[position].actor.visible = visible;
+    }
+};
+
+function PopupComboBoxMenuItem() {
+    this._init.apply(this, arguments);
+}
+
+PopupComboBoxMenuItem.prototype = {
+    __proto__: PopupBaseMenuItem.prototype,
+
+    _init: function (params) {
+        PopupBaseMenuItem.prototype._init.call(this, params);
+
+        this._itemBox = new Shell.Stack();
+        this.addActor(this._itemBox);
+
+        let expander = new St.Label({ text: '\u2304' });
+        this.addActor(expander, { align: St.Align.END });
+
+        this._menu = new PopupComboMenu(this.actor);
+        Main.uiGroup.add_actor(this._menu.actor);
+        this._menu.actor.hide();
+
+        if (params.style_class)
+            this._menu.actor.add_style_class_name(params.style_class);
+
+        this._activeItemPos = -1;
+        this._items = [];
+    },
+
+    _getTopMenu: function() {
+        let actor = this.actor.get_parent();
+        while (actor) {
+            if (actor._delegate &&
+                (actor._delegate instanceof PopupMenu ||
+                 actor._delegate instanceof PopupComboMenu))
+                return actor._delegate;
+
+            actor = actor.get_parent();
+        }
+
+        return null;
+    },
+
+    activate: function(event) {
+        let topMenu = this._getTopMenu();
+        if (!topMenu)
+            return;
+
+        topMenu.addChildMenu(this._menu);
+        this._menu.toggle();
+    },
+
+    addMenuItem: function(menuItem, position) {
+        if (position === undefined)
+            position = this._menu.numMenuItems;
+
+        this._menu.addMenuItem(menuItem, position);
+        _ensureStyle(this._menu.actor);
+
+        let item = new St.BoxLayout({ style_class: 'popup-combobox-item' });
+
+        let children = menuItem.actor.get_children();
+        for (let i = 0; i < children.length; i++) {
+            let clone = new Clutter.Clone({ source: children[i] });
+            item.add(clone, { y_fill: false });
+        }
+
+        let oldItem = this._items[position];
+        if (oldItem)
+            this._itemBox.remove_actor(oldItem);
+
+        this._items[position] = item;
+        this._itemBox.add_actor(item);
+
+        menuItem.connect('activate',
+                         Lang.bind(this, this._itemActivated, position));
+    },
+
+    setActiveItem: function(position) {
+        let item = this._items[position];
+        if (!item)
+            return;
+        if (this._activeItemPos == position)
+            return;
+        this._menu.setActiveItem(position);
+        this._activeItemPos = position;
+        for (let i = 0; i < this._items.length; i++)
+            this._items[i].visible = (i == this._activeItemPos);
+    },
+
+    setItemVisible: function(position, visible) {
+        this._menu.setItemVisible(position, visible);
+    },
+
+    _itemActivated: function(menuItem, event, position) {
+        this.setActiveItem(position);
+        this.emit('active-item-changed', position);
+    }
+};
 
 /* Basic implementation of a menu manager.
  * Call addMenu to add menus
@@ -1337,6 +1626,7 @@ PopupMenuManager.prototype = {
         this._keyFocusNotifyId = 0;
         this._activeMenu = null;
         this._menus = [];
+        this._menuStack = [];
         this._preGrabInputMode = null;
         this._grabbedFromKeynav = false;
     },
@@ -1345,6 +1635,8 @@ PopupMenuManager.prototype = {
         let menudata = {
             menu:              menu,
             openStateChangeId: menu.connect('open-state-changed', Lang.bind(this, this._onMenuOpenState)),
+            childMenuAddedId:  menu.connect('child-menu-added', Lang.bind(this, this._onChildMenuAdded)),
+            childMenuRemovedId: menu.connect('child-menu-removed', Lang.bind(this, this._onChildMenuRemoved)),
             destroyId:         menu.connect('destroy', Lang.bind(this, this._onMenuDestroy)),
             enterId:           0,
             focusInId:         0
@@ -1372,6 +1664,8 @@ PopupMenuManager.prototype = {
 
         let menudata = this._menus[position];
         menu.disconnect(menudata.openStateChangeId);
+        menu.disconnect(menudata.childMenuAddedId);
+        menu.disconnect(menudata.childMenuRemovedId);
         menu.disconnect(menudata.destroyId);
 
         if (menudata.enterId)
@@ -1409,8 +1703,20 @@ PopupMenuManager.prototype = {
     },
 
     _onMenuOpenState: function(menu, open) {
-        if (open)
+        if (open) {
+            if (this._activeMenu && this._activeMenu.isChildMenu(menu)) {
+                this._menuStack.push(this._activeMenu);
+                menu.actor.grab_key_focus();
+            }
             this._activeMenu = menu;
+        } else {
+            if (this._menuStack.length > 0) {
+                this._activeMenu = this._menuStack.pop();
+                if (menu.sourceActor)
+                    menu.sourceActor.grab_key_focus();
+                this._didPop = true;
+            }
+        }
 
         // Check what the focus was before calling pushModal/popModal
         let focus = global.stage.key_focus;
@@ -1443,6 +1749,14 @@ PopupMenuManager.prototype = {
         }
     },
 
+    _onChildMenuAdded: function(menu, childMenu) {
+        this.addMenu(childMenu);
+    },
+
+    _onChildMenuRemoved: function(menu, childMenu) {
+        this.removeMenu(childMenu);
+    },
+
     // change the currently-open menu without dropping grab
     _changeMenu: function(newMenu) {
         if (this._activeMenu) {
@@ -1451,6 +1765,8 @@ PopupMenuManager.prototype = {
             // before closing it to keep that from happening
             let oldMenu = this._activeMenu;
             this._activeMenu = null;
+            for (let i = this._menuStack.length - 1; i >= 0; i--)
+                this._menuStack[i].close(false);
             oldMenu.close(false);
             newMenu.open(false);
         } else
@@ -1461,6 +1777,15 @@ PopupMenuManager.prototype = {
         if (!this.grabbed || menu == this._activeMenu)
             return false;
 
+        if (this._activeMenu && this._activeMenu.isChildMenu(menu))
+            return false;
+
+        if (this._menuStack.indexOf(menu) != -1)
+            return false;
+
+        if (this._menuStack.length > 0 && this._menuStack[0].isChildMenu(menu))
+            return false;
+
         this._changeMenu(menu);
         return false;
     },
@@ -1473,6 +1798,8 @@ PopupMenuManager.prototype = {
         if (focus) {
             if (this._activeMenuContains(focus))
                 return;
+            if (this._menuStack.length > 0)
+                return;
             if (focus._delegate && focus._delegate.menu &&
                 this._findMenu(focus._delegate.menu) != -1)
                 return;
@@ -1531,6 +1858,11 @@ PopupMenuManager.prototype = {
         if (this._activeMenu != null && this._activeMenu.passEvents)
             return false;
 
+        if (this._didPop) {
+            this._didPop = false;
+            return true;
+        }
+
         let activeMenuContains = this._eventIsOnActiveMenu(event);
         let eventType = event.type();
 
