@@ -1,6 +1,7 @@
 /* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */
 
 const DBus = imports.dbus;
+const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
 const Lang = imports.lang;
 const Mainloop = imports.mainloop;
@@ -11,6 +12,7 @@ const Tpl = imports.gi.TelepathyLogger;
 const Tp = imports.gi.TelepathyGLib;
 
 const History = imports.misc.history;
+const Params = imports.misc.params;
 const Main = imports.ui.main;
 const MessageTray = imports.ui.messageTray;
 
@@ -43,11 +45,16 @@ let contactFeatures = [Tp.ContactFeature.ALIAS,
 
 function makeMessageFromTpMessage(tpMessage, direction) {
     let [text, flags] = tpMessage.to_text();
+
+    let timestamp = tpMessage.get_sent_timestamp();
+    if (timestamp == 0)
+        timestamp = tpMessage.get_received_timestamp();
+
     return {
         messageType: tpMessage.get_message_type(),
         text: text,
         sender: tpMessage.sender.alias,
-        timestamp: tpMessage.get_received_timestamp(),
+        timestamp: timestamp,
         direction: direction
     };
 }
@@ -80,8 +87,8 @@ Client.prototype = {
         // channel matching its filters is detected.
         // The second argument, recover, means _observeChannels will be run
         // for any existing channel as well.
-        let dbus = Tp.DBusDaemon.dup();
-        this._tpClient = new Shell.TpClient({ 'dbus_daemon': dbus,
+        this._accountManager = Tp.AccountManager.dup();
+        this._tpClient = new Shell.TpClient({ 'account-manager': this._accountManager,
                                               'name': 'GnomeShell',
                                               'uniquify-name': true })
         this._tpClient.set_observe_channels_func(
@@ -91,6 +98,11 @@ Client.prototype = {
         this._tpClient.set_handle_channels_func(
             Lang.bind(this, this._handleChannels));
 
+        // Workaround for gjs not supporting GPtrArray in signals.
+        // See BGO bug #653941 for context.
+        this._tpClient.set_contact_list_changed_func(
+            Lang.bind(this, this._contactListChanged));
+
         // Allow other clients (such as Empathy) to pre-empt our channels if
         // needed
         this._tpClient.set_delegated_channels_callback(
@@ -101,6 +113,22 @@ Client.prototype = {
         } catch (e) {
             throw new Error('Couldn\'t register Telepathy client. Error: \n' + e);
         }
+
+
+        // Watch subscription requests and connection errors
+        this._subscriptionSource = null;
+        this._accountSource = null;
+        let factory = this._accountManager.get_factory();
+        factory.add_account_features([Tp.Account.get_feature_quark_connection()]);
+        factory.add_connection_features([Tp.Connection.get_feature_quark_contact_list()]);
+        factory.add_contact_features([Tp.ContactFeature.SUBSCRIPTION_STATES,
+                                      Tp.ContactFeature.ALIAS,
+                                      Tp.ContactFeature.AVATAR_DATA]);
+
+        this._accountManager.connect('account-validity-changed',
+            Lang.bind(this, this._accountValidityChanged));
+
+        this._accountManager.prepare_async(null, Lang.bind(this, this._accountManagerPrepared));
     },
 
     _observeChannels: function(observer, account, conn, channels,
@@ -225,7 +253,8 @@ Client.prototype = {
 
         // FIXME: We don't have a 'chat room' icon (bgo #653737) use
         // system-users for now as Empathy does.
-        let source = new ApproverSource(dispatchOp, _("Invitation"), 'system-users');
+        let source = new ApproverSource(dispatchOp, _("Invitation"),
+                                        Shell.util_icon_from_string('system-users'));
         Main.messageTray.add(source);
 
         let notif = new RoomInviteNotification(source, dispatchOp, channel, contacts[0]);
@@ -243,6 +272,8 @@ Client.prototype = {
         else if (chanType == Tp.IFACE_CHANNEL_TYPE_STREAMED_MEDIA ||
                  chanType == 'org.freedesktop.Telepathy.Channel.Type.Call.DRAFT')
             this._approveCall(account, conn, channel, dispatchOp, context);
+        else if (chanType == Tp.IFACE_CHANNEL_TYPE_FILE_TRANSFER)
+            this._approveFileTransfer(account, conn, channel, dispatchOp, context);
     },
 
     _approveTextChannel: function(account, conn, channel, dispatchOp, context) {
@@ -288,7 +319,9 @@ Client.prototype = {
           isVideo = true;
 
         // We got the TpContact
-        let source = new ApproverSource(dispatchOp, _("Call"), isVideo ? 'camera-web' : 'audio-input-microphone');
+        let source = new ApproverSource(dispatchOp, _("Call"), isVideo ?
+                                        Shell.util_icon_from_string('camera-web') :
+                                        Shell.util_icon_from_string('audio-input-microphone'));
         Main.messageTray.add(source);
 
         let notif = new AudioVideoNotification(source, dispatchOp, channel, contacts[0], isVideo);
@@ -296,9 +329,141 @@ Client.prototype = {
         context.accept();
     },
 
+    _approveFileTransfer: function(account, conn, channel, dispatchOp, context) {
+        let [targetHandle, targetHandleType] = channel.get_handle();
+
+        Shell.get_tp_contacts(conn, [targetHandle],
+                contactFeatures,
+                Lang.bind(this, this._createFileTransferSource, channel, context, dispatchOp));
+    },
+
+    _createFileTransferSource: function(connection, contacts, failed, channel, context, dispatchOp) {
+        if (contacts.length < 1) {
+            Shell.decline_dispatch_op(context, 'Failed to get file sender');
+            return;
+        }
+
+        // Use the icon of the file being transferred
+        let gicon = Gio.content_type_get_icon(channel.get_mime_type());
+
+        // We got the TpContact
+        let source = new ApproverSource(dispatchOp, _("File Transfer"), gicon);
+        Main.messageTray.add(source);
+
+        let notif = new FileTransferNotification(source, dispatchOp, channel, contacts[0]);
+        source.notify(notif);
+        context.accept();
+    },
+
     _delegatedChannelsCb: function(client, channels) {
         // Nothing to do as we don't make a distinction between observed and
         // handled channels.
+    },
+
+    _accountManagerPrepared: function(am, result) {
+        am.prepare_finish(result);
+
+        let accounts = am.get_valid_accounts();
+        for (let i = 0; i < accounts.length; i++) {
+            this._accountValidityChanged(am, accounts[i], true);
+        }
+    },
+
+    _accountValidityChanged: function(am, account, valid) {
+        if (!valid)
+            return;
+
+        // It would be better to connect to "status-changed" but we cannot.
+        // See discussion in https://bugzilla.gnome.org/show_bug.cgi?id=654159
+        account.connect("notify::connection-status",
+                        Lang.bind(this, this._accountConnectionStatusNotifyCb));
+
+        account.connect('notify::connection',
+                        Lang.bind(this, this._connectionChanged));
+        this._connectionChanged(account);
+    },
+
+    _connectionChanged: function(account) {
+        let conn = account.get_connection();
+        if (conn == null)
+            return;
+
+        this._tpClient.grab_contact_list_changed(conn);
+        if (conn.get_contact_list_state() == Tp.ContactListState.SUCCESS) {
+            this._contactListChanged(conn, conn.dup_contact_list(), []);
+        }
+    },
+
+    _contactListChanged: function(conn, added, removed) {
+        for (let i = 0; i < added.length; i++) {
+            let contact = added[i];
+
+            contact.connect('subscription-states-changed',
+                            Lang.bind(this, this._subscriptionStateChanged));
+            this._subscriptionStateChanged(contact);
+        }
+    },
+
+    _subscriptionStateChanged: function(contact) {
+        if (contact.get_publish_state() != Tp.SubscriptionState.ASK)
+            return;
+
+        /* Implicitly accept publish requests if contact is already subscribed */
+        if (contact.get_subscribe_state() == Tp.SubscriptionState.YES ||
+            contact.get_subscribe_state() == Tp.SubscriptionState.ASK) {
+
+            contact.authorize_publication_async(function(src, result) {
+                src.authorize_publication_finish(result)});
+
+            return;
+        }
+
+        /* Display notification to ask user to accept/reject request */
+        let source = this._ensureSubscriptionSource();
+        Main.messageTray.add(source);
+
+        let notif = new SubscriptionRequestNotification(source, contact);
+        source.notify(notif);
+    },
+
+    _ensureSubscriptionSource: function() {
+        if (this._subscriptionSource == null) {
+            this._subscriptionSource = new MultiNotificationSource(
+                _("Subscription request"), 'gtk-dialog-question');
+            this._subscriptionSource.connect('destroy', Lang.bind(this, function () {
+                this._subscriptionSource = null;
+            }));
+        }
+
+        return this._subscriptionSource;
+    },
+
+    _accountConnectionStatusNotifyCb: function(account) {
+        let connectionError = account.connection_error;
+
+        if (account.connection_status != Tp.ConnectionStatus.DISCONNECTED ||
+            connectionError == Tp.error_get_dbus_name(Tp.Error.CANCELLED)) {
+            return;
+        }
+
+        /* Display notification that account failed to connect */
+        let source = this._ensureAccountSource();
+        Main.messageTray.add(source);
+
+        let notif = new AccountNotification(source, account, connectionError);
+        source.notify(notif);
+    },
+
+    _ensureAccountSource: function() {
+        if (this._accountSource == null) {
+            this._accountSource = new MultiNotificationSource(
+                _("Connection error"), 'gtk-dialog-error');
+            this._accountSource.connect('destroy', Lang.bind(this, function () {
+                this._accountSource = null;
+            }));
+        }
+
+        return this._accountSource;
     }
 };
 
@@ -326,6 +491,7 @@ ChatSource.prototype = {
 
         this._notification = new ChatNotification(this);
         this._notification.setUrgency(MessageTray.Urgency.HIGH);
+        this._notifyTimeoutId = 0;
 
         // We ack messages when the message box is collapsed if user has
         // interacted with it before and so read the messages:
@@ -487,7 +653,22 @@ ChatSource.prototype = {
 
         message = makeMessageFromTpMessage(message, NotificationDirection.RECEIVED);
         this._notification.appendMessage(message);
-        this.notify();
+
+        // Wait a bit before notifying for the received message, a handler
+        // could ack it in the meantime.
+        if (this._notifyTimeoutId != 0)
+            Mainloop.source_remove(this._notifyTimeoutId);
+        this._notifyTimeoutId = Mainloop.timeout_add(500,
+            Lang.bind(this, this._notifyTimeout));
+    },
+
+    _notifyTimeout: function() {
+        if (this._pendingMessages.length != 0)
+            this.notify();
+
+        this._notifyTimeoutId = 0;
+
+        return false;
     },
 
     // This is called for both messages we send from
@@ -576,13 +757,10 @@ ChatSource.prototype = {
     },
 
     _ackMessages: function() {
-        if (this._pendingMessages.length == 0)
-            return;
-
         // Don't clear our messages here, tp-glib will send a
         // 'pending-message-removed' for each one.
-        this._channel.ack_messages_async(this._pendingMessages, Lang.bind(this, function(src, result) {
-            this._channel.ack_messages_finish(result);}));
+        this._channel.ack_all_pending_messages_async(Lang.bind(this, function(src, result) {
+            this._channel.ack_all_pending_messages_finish(result);}));
     },
 
     _summaryItemClicked: function(source, button) {
@@ -623,6 +801,8 @@ ChatNotification.prototype = {
 
         this._oldMaxScrollAdjustment = 0;
         this._createScrollArea();
+        this._lastGroup = null;
+        this._lastGroupActor = null;
 
         this._scrollArea.vscroll.adjustment.connect('changed', Lang.bind(this, function(adjustment) {
             let currentValue = adjustment.value + adjustment.page_size;
@@ -650,12 +830,10 @@ ChatNotification.prototype = {
      * @noTimestamp: Whether to add a timestamp. If %true, no timestamp
      *   will be added, regardless of the difference since the
      *   last timestamp
-     * @styles: A list of CSS class names.
      */
-    appendMessage: function(message, noTimestamp, styles) {
+    appendMessage: function(message, noTimestamp) {
         let messageBody = GLib.markup_escape_text(message.text, -1);
-        styles = styles || [];
-        styles.push(message.direction);
+        let styles = [message.direction];
 
         if (message.messageType == Tp.ChannelTextMessageType.ACTION) {
             let senderAlias = GLib.markup_escape_text(message.sender, -1);
@@ -668,7 +846,14 @@ ChatNotification.prototype = {
                                                           bannerMarkup: true });
         }
 
-        this._append(messageBody, styles, message.timestamp, noTimestamp);
+        let group = (message.direction == NotificationDirection.RECEIVED ?
+                     'received' : 'sent');
+
+        this._append({ body: messageBody,
+                       group: group,
+                       styles: styles,
+                       timestamp: message.timestamp,
+                       noTimestamp: noTimestamp });
     },
 
     _filterMessages: function() {
@@ -694,24 +879,65 @@ ChatNotification.prototype = {
             for (let i = 0; i < expired.length; i++)
                 expired[i].actor.destroy();
         }
+
+        let groups = this._contentArea.get_children();
+        for (let i = 0; i < groups.length; i ++) {
+            let group = groups[i];
+            if (group.get_children().length == 0)
+                group.destroy();
+        }
     },
 
-    _append: function(text, styles, timestamp, noTimestamp) {
+    /**
+     * _append:
+     * @props: An object with the properties:
+     *  body: The text of the message.
+     *  group: The group of the message, one of:
+     *         'received', 'sent', 'meta'.
+     *  styles: Style class names for the message to have.
+     *  timestamp: The timestamp of the message.
+     *  noTimestamp: suppress timestamp signal?
+     *  childProps: props to add the actor with.
+     */
+    _append: function(props) {
         let currentTime = (Date.now() / 1000);
-        if (!timestamp)
-            timestamp = currentTime;
+        props = Params.parse(props, { body: null,
+                                      group: null,
+                                      styles: [],
+                                      timestamp: currentTime,
+                                      noTimestamp: false,
+                                      childProps: null });
 
         // Reset the old message timeout
         if (this._timestampTimeoutId)
             Mainloop.source_remove(this._timestampTimeoutId);
 
-        let body = this.addBody(text, true);
+        let highlighter = new MessageTray.URLHighlighter(props.body,
+                                                         true,  // line wrap?
+                                                         true); // allow markup?
+
+        let body = highlighter.actor;
+
+        let styles = props.styles;
         for (let i = 0; i < styles.length; i ++)
             body.add_style_class_name(styles[i]);
 
-        this._history.unshift({ actor: body, time: timestamp, realMessage: true });
+        let group = props.group;
+        if (group != this._lastGroup) {
+            let style = 'chat-group-' + group;
+            this._lastGroup = group;
+            this._lastGroupActor = new St.BoxLayout({ style_class: style,
+                                                      vertical: true });
+            this.addActor(this._lastGroupActor);
+        }
 
-        if (!noTimestamp) {
+        this._lastGroupActor.add(body, props.childProps);
+
+        let timestamp = props.timestamp;
+        this._history.unshift({ actor: body, time: timestamp,
+                                realMessage: group != 'meta' });
+
+        if (!props.noTimestamp) {
             if (timestamp < currentTime - SCROLLBACK_IMMEDIATE_TIME)
                 this.appendTimestamp();
             else
@@ -738,18 +964,18 @@ ChatNotification.prototype = {
              If applicable, replace %X with a strftime format valid for your
              locale, without seconds. */
             // xgettext:no-c-format
-            format = _("Sent at %X on %A");
-
-        // FIXME: The next two are stolen from calendar.js with the comment to avoid
-        // a string-freeze break. They should be replaced with better strings
-        // with 'Sent at', appropriate context and appropriate translator comment.
+            format = _("Sent at <b>%X</b> on <b>%A</b>");
 
         } else if (date.getYear() == now.getYear()) {
-            /* Translators: Shown on calendar heading when selected day occurs on current year */
-            format = C_("calendar heading", "%A, %B %d");
+            /* Translators: this is a time format in the style of "Wednesday, May 25",
+             shown when you get a chat message in the same year. */
+            // xgettext:no-c-format
+            format = _("Sent on <b>%A</b>, <b>%B %d</b>");
         } else {
-            /* Translators: Shown on calendar heading when selected day occurs on different year */
-            format = C_("calendar heading", "%A, %B %d, %Y");
+            /* Translators: this is a time format in the style of "Wednesday, May 25, 2012",
+             shown when you get a chat message in a different year. */
+            // xgettext:no-c-format
+            format = _("Sent on <b>%A</b>, <b>%B %d</b>, %Y");
         }
 
         return date.toLocaleFormat(format);
@@ -759,11 +985,13 @@ ChatNotification.prototype = {
         let lastMessageTime = this._history[0].time;
         let lastMessageDate = new Date(lastMessageTime * 1000);
 
-        let timeLabel = this.addBody(this._formatTimestamp(lastMessageDate), false, { expand: true, x_fill: false, x_align: St.Align.END });
-        timeLabel.add_style_class_name('chat-meta-message');
-        this._history.unshift({ actor: timeLabel, time: lastMessageTime, realMessage: false });
-
-        this._timestampTimeoutId = 0;
+        let timeLabel = this._append({ body: this._formatTimestamp(lastMessageDate),
+                                       group: 'meta',
+                                       styles: ['chat-meta-message'],
+                                       childProps: { expand: true, x_fill: false,
+                                                     x_align: St.Align.END },
+                                       noTimestamp: true,
+                                       timestamp: lastMessageTime });
 
         this._filterMessages();
 
@@ -775,9 +1003,10 @@ ChatNotification.prototype = {
             this.update(text, null, { customContent: true, titleMarkup: true });
         else
             this.update(this.source.title, null, { customContent: true });
-        let label = this.addBody(text, true);
-        label.add_style_class_name('chat-meta-message');
-        this._history.unshift({ actor: label, time: (Date.now() / 1000), realMessage: false});
+
+        let label = this._append({ body: text,
+                                   group: 'meta',
+                                   styles: ['chat-meta-message'] });
 
         this._filterMessages();
     },
@@ -789,9 +1018,11 @@ ChatNotification.prototype = {
         /* Translators: this is the other person changing their old IM name to their new
            IM name. */
         let message = '<i>' + _("%s is now known as %s").format(oldAlias, newAlias) + '</i>';
-        let label = this.addBody(message, true);
-        label.add_style_class_name('chat-meta-message');
-        this._history.unshift({ actor: label, time: (Date.now() / 1000), realMessage: false });
+
+        let label = this._append({ body: text,
+                                   group: 'meta',
+                                   styles: ['chat-meta-message'] });
+
         this.update(newAlias, null, { customContent: true });
 
         this._filterMessages();
@@ -844,17 +1075,17 @@ ChatNotification.prototype = {
     }
 };
 
-function ApproverSource(dispatchOp, text, icon) {
-    this._init(dispatchOp, text, icon);
+function ApproverSource(dispatchOp, text, gicon) {
+    this._init(dispatchOp, text, gicon);
 }
 
 ApproverSource.prototype = {
     __proto__: MessageTray.Source.prototype,
 
-    _init: function(dispatchOp, text, icon) {
+    _init: function(dispatchOp, text, gicon) {
         MessageTray.Source.prototype._init.call(this, text);
 
-        this._icon = icon;
+        this._gicon = gicon;
         this._setSummaryIcon(this.createNotificationIcon());
 
         this._dispatchOp = dispatchOp;
@@ -877,7 +1108,7 @@ ApproverSource.prototype = {
     },
 
     createNotificationIcon: function() {
-        return new St.Icon({ icon_name: this._icon,
+        return new St.Icon({ gicon: this._gicon,
                              icon_type: St.IconType.FULLCOLOR,
                              icon_size: this.ICON_SIZE });
     }
@@ -971,3 +1202,321 @@ AudioVideoNotification.prototype = {
         }));
     }
 };
+
+// File Transfer
+function FileTransferNotification(source, dispatchOp, channel, contact) {
+    this._init(source, dispatchOp, channel, contact);
+}
+
+FileTransferNotification.prototype = {
+    __proto__: MessageTray.Notification.prototype,
+
+    _init: function(source, dispatchOp, channel, contact) {
+        MessageTray.Notification.prototype._init.call(this,
+                                                      source,
+                                                      /* To translators: The first parameter is
+                                                       * the contact's alias and the second one is the
+                                                       * file name. The string will be something
+                                                       * like: "Alice is sending you test.ogg"
+                                                       */
+                                                      _("%s is sending you %s").format(contact.get_alias(),
+                                                                                       channel.get_filename()),
+                                                      null,
+                                                      { customContent: true });
+        this.setResident(true);
+
+        this.addButton('decline', _("Decline"));
+        this.addButton('accept', _("Accept"));
+
+        this.connect('action-invoked', Lang.bind(this, function(self, action) {
+            switch (action) {
+            case 'decline':
+                dispatchOp.leave_channels_async(Tp.ChannelGroupChangeReason.NONE,
+                                                '', function(src, result) {
+                    src.leave_channels_finish(result)});
+                break;
+            case 'accept':
+                dispatchOp.handle_with_time_async('', global.get_current_time(),
+                                                  function(src, result) {
+                    src.handle_with_time_finish(result)});
+                break;
+            }
+            this.destroy();
+        }));
+    }
+};
+
+// A notification source that can embed multiple notifications
+function MultiNotificationSource(title, icon) {
+    this._init(title, icon);
+}
+
+MultiNotificationSource.prototype = {
+    __proto__: MessageTray.Source.prototype,
+
+    _init: function(title, icon) {
+        MessageTray.Source.prototype._init.call(this, title);
+
+        this._icon = icon;
+        this._setSummaryIcon(this.createNotificationIcon());
+        this._nbNotifications = 0;
+    },
+
+    notify: function(notification) {
+        MessageTray.Source.prototype.notify.call(this, notification);
+
+        this._nbNotifications += 1;
+
+        // Display the source while there is at least one notification
+        notification.connect('destroy', Lang.bind(this, function () {
+            this._nbNotifications -= 1;
+
+            if (this._nbNotifications == 0)
+                this.destroy();
+        }));
+    },
+
+    createNotificationIcon: function() {
+        return new St.Icon({ gicon: Shell.util_icon_from_string(this._icon),
+                             icon_type: St.IconType.FULLCOLOR,
+                             icon_size: this.ICON_SIZE });
+    }
+};
+
+// Subscription request
+function SubscriptionRequestNotification(source, contact) {
+    this._init(source, contact);
+}
+
+SubscriptionRequestNotification.prototype = {
+    __proto__: MessageTray.Notification.prototype,
+
+    _init: function(source, contact) {
+        MessageTray.Notification.prototype._init.call(this, source,
+            /* To translators: The parameter is the contact's alias */
+            _("%s would like permission to see when you are online").format(contact.get_alias()),
+            null, { customContent: true });
+
+        this._contact = contact;
+        this._connection = contact.get_connection();
+
+        let layout = new St.BoxLayout({ vertical: false });
+
+        // Display avatar
+        let iconBox = new St.Bin({ style_class: 'avatar-box' });
+        iconBox._size = 48;
+
+        let textureCache = St.TextureCache.get_default();
+        let file = contact.get_avatar_file();
+
+        if (file) {
+            let uri = file.get_uri();
+            iconBox.child = textureCache.load_uri_async(uri, iconBox._size, iconBox._size);
+        }
+        else {
+            iconBox.child = new St.Icon({ icon_name: 'avatar-default',
+                                          icon_type: St.IconType.FULLCOLOR,
+                                          icon_size: iconBox._size });
+        }
+
+        layout.add(iconBox);
+
+        // subscription request message
+        let label = new St.Label({ style_class: 'subscription-message',
+                                   text: contact.get_publish_request() });
+
+        layout.add(label);
+
+        this.addActor(layout);
+
+        this.addButton('decline', _("Decline"));
+        this.addButton('accept', _("Accept"));
+
+        this.connect('action-invoked', Lang.bind(this, function(self, action) {
+            switch (action) {
+            case 'decline':
+                contact.remove_async(function(src, result) {
+                    src.remove_finish(result)});
+                break;
+            case 'accept':
+                // Authorize the contact and request to see his status as well
+                contact.authorize_publication_async(function(src, result) {
+                    src.authorize_publication_finish(result)});
+
+                contact.request_subscription_async('', function(src, result) {
+                    src.request_subscription_finish(result)});
+                break;
+            }
+
+            // rely on _subscriptionStatesChangedCb to destroy the
+            // notification
+        }));
+
+        this._changedId = contact.connect('subscription-states-changed',
+            Lang.bind(this, this._subscriptionStatesChangedCb));
+        this._invalidatedId = this._connection.connect('invalidated',
+            Lang.bind(this, this.destroy));
+    },
+
+    destroy: function() {
+        if (this._changedId != 0) {
+            this._contact.disconnect(this._changedId);
+            this._changedId = 0;
+        }
+
+        if (this._invalidatedId != 0) {
+            this._connection.disconnect(this._invalidatedId);
+            this._invalidatedId = 0;
+        }
+
+        MessageTray.Notification.prototype.destroy.call(this);
+    },
+
+    _subscriptionStatesChangedCb: function(contact, subscribe, publish, msg) {
+        // Destroy the notification if the subscription request has been
+        // answered
+        if (publish != Tp.SubscriptionState.ASK)
+            this.destroy();
+    }
+};
+
+
+function AccountNotification(source, account, connectionError) {
+    this._init(source, account, connectionError);
+}
+
+// Messages from empathy/libempathy/empathy-utils.c
+// create_errors_to_message_hash()
+
+/* Translator note: these should be the same messages that are
+ * used in Empathy, so just copy and paste from there. */
+let _connectionErrorMessages = {};
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.NETWORK_ERROR)]
+  = _("Network error");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.AUTHENTICATION_FAILED)]
+  = _("Authentication failed");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.ENCRYPTION_ERROR)]
+  = _("Encryption error");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_NOT_PROVIDED)]
+  = _("Certificate not provided");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_UNTRUSTED)]
+  = _("Certificate untrusted");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_EXPIRED)]
+  = _("Certificate expired");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_NOT_ACTIVATED)]
+  = _("Certificate not activated");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_HOSTNAME_MISMATCH)]
+  = _("Certificate hostname mismatch");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_FINGERPRINT_MISMATCH)]
+  = _("Certificate fingerprint mismatch");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_SELF_SIGNED)]
+  = _("Certificate self-signed");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CANCELLED)]
+  = _("Status is set to offline");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.ENCRYPTION_NOT_AVAILABLE)]
+  = _("Encryption is not available");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_INVALID)]
+  = _("Certificate is invalid");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CONNECTION_REFUSED)]
+  = _("Connection has been refused");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CONNECTION_FAILED)]
+  = _("Connection can't be established");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CONNECTION_LOST)]
+  = _("Connection has been lost");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.ALREADY_CONNECTED)]
+  = _("This resource is already connected to the server");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CONNECTION_REPLACED)]
+  = _("Connection has been replaced by a new connection using the "
+    + "same resource");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.REGISTRATION_EXISTS)]
+  = _("The account already exists on the server");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.SERVICE_BUSY)]
+  = _("Server is currently too busy to handle the connection");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_REVOKED)]
+  = _("Certificate has been revoked");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_INSECURE)]
+  = _("Certificate uses an insecure cipher algorithm or is "
+    + "cryptographically weak");
+_connectionErrorMessages[Tp.error_get_dbus_name(Tp.Error.CERT_LIMIT_EXCEEDED)]
+  = _("The length of the server certificate, or the depth of the "
+    + "server certificate chain, exceed the limits imposed by the "
+    + "cryptography library");
+
+AccountNotification.prototype = {
+    __proto__: MessageTray.Notification.prototype,
+
+    _init: function(source, account, connectionError) {
+        MessageTray.Notification.prototype._init.call(this, source,
+            /* translators: argument is the account name, like
+             * name@jabber.org for example. */
+            _("Connection to %s failed").format(account.get_display_name()),
+            null, { customContent: true });
+
+        let message;
+        if (connectionError in _connectionErrorMessages) {
+            message = _connectionErrorMessages[connectionError];
+        } else {
+            message = _("Unknown reason");
+        }
+
+        this._account = account;
+
+        this.addBody(message);
+
+        this.addButton('reconnect', _("Reconnect"));
+        this.addButton('edit', _("Edit account"));
+
+        this.connect('action-invoked', Lang.bind(this, function(self, action) {
+            switch (action) {
+            case 'reconnect':
+                // If it fails again, a new notification should pop up with the
+                // new error.
+                account.reconnect_async(null, null);
+                break;
+            case 'edit':
+                let cmd = '/usr/bin/empathy-accounts'
+                        + ' --select-account=%s'
+                        .format(account.get_path_suffix());
+                let app_info = Gio.app_info_create_from_commandline(cmd, null, 0,
+                    null);
+                app_info.launch([], null, null);
+                break;
+            }
+            this.destroy();
+        }));
+
+        this._enabledId = account.connect('notify::enabled',
+                                          Lang.bind(this, function() {
+                                              if (!account.is_enabled())
+                                                  this.destroy();
+                                          }));
+
+        this._invalidatedId = account.connect('invalidated',
+                                              Lang.bind(this, this.destroy));
+
+        this._connectionStatusId = account.connect('notify::connection-status',
+            Lang.bind(this, function() {
+                if (account.connection_status != Tp.ConnectionStatus.DISCONNECTED)
+                    this.destroy();
+            }));
+    },
+
+    destroy: function() {
+        if (this._enabledId != 0) {
+            this._account.disconnect(this._enabledId);
+            this._enabledId = 0;
+        }
+
+        if (this._invalidatedId != 0) {
+            this._account.disconnect(this._invalidatedId);
+            this._invalidatedId = 0;
+        }
+
+        if (this._connectionStatusId != 0) {
+            this._account.disconnect(this._connectionStatusId);
+            this._connectionStatusId = 0;
+        }
+
+        MessageTray.Notification.prototype.destroy.call(this);
+    }
+};
