@@ -31,6 +31,7 @@ ShellGlobal   *shell_global_get                       (void);
 
 MetaScreen    *shell_global_get_screen                (ShellGlobal *global);
 GdkScreen     *shell_global_get_gdk_screen            (ShellGlobal *global);
+MetaDisplay   *shell_global_get_display               (ShellGlobal *global);
 GList         *shell_global_get_window_actors         (ShellGlobal *global);
 GSettings     *shell_global_get_settings              (ShellGlobal *global);
 guint32        shell_global_get_current_time          (ShellGlobal *global);
@@ -141,6 +142,30 @@ void     shell_global_reexec_self               (ShellGlobal  *global);
 
 void     shell_global_launch_calendar_server    (ShellGlobal  *global);
 
+typedef void (*ShellGlobalScreenshotCallback)  (ShellGlobal *global, gboolean success);
+
+void    shell_global_screenshot_area           (ShellGlobal  *global,
+                                                int x,
+                                                int y,
+                                                int width,
+                                                int height,
+                                                const char *filename,
+                                                ShellGlobalScreenshotCallback callback);
+
+gboolean shell_global_screenshot_window         (ShellGlobal  *global,
+                                                gboolean include_frame,
+                                                const char *filename);
+
+void    shell_global_screenshot                (ShellGlobal  *global,
+                                                const char *filename,
+                                                ShellGlobalScreenshotCallback callback);
+typedef enum {
+  SHELL_SESSION_USER,
+  SHELL_SESSION_GDM
+} ShellSessionType;
+
+ShellSessionType shell_global_get_session_type  (ShellGlobal  *global);
+
 G_END_DECLS
 
 #endif /* __SHELL_GLOBAL_H__ */
