@@ -18,7 +18,6 @@
 #include <libsn/sn.h>
 
 #include "shell-window-tracker-private.h"
-#include "shell-app-system.h"
 #include "shell-app-private.h"
 #include "shell-global.h"
 #include "shell-marshal.h"
@@ -41,24 +40,6 @@
  * have it also track through startup-notification.
  */
 
-/* Title patterns to detect apps that don't set WM class as needed.
- * Format: application ID, title regex pattern, NULL (for GRegex) */
-static struct
-{
-  const char *app_id;
-  const char *pattern;
-  GRegex *regex;
-} title_patterns[] =  {
-    {"mozilla-firefox.desktop", ".* - Mozilla Firefox", NULL}, \
-    {"openoffice.org-writer.desktop", ".* - OpenOffice.org Writer$", NULL}, \
-    {"openoffice.org-calc.desktop", ".* - OpenOffice.org Calc$", NULL}, \
-    {"openoffice.org-impress.desktop", ".* - OpenOffice.org Impress$", NULL}, \
-    {"openoffice.org-draw.desktop", ".* - OpenOffice.org Draw$", NULL}, \
-    {"openoffice.org-base.desktop", ".* - OpenOffice.org Base$", NULL}, \
-    {"openoffice.org-math.desktop", ".* - OpenOffice.org Math$", NULL}, \
-    {NULL, NULL, NULL}
-};
-
 struct _ShellWindowTracker
 {
   GObject parent;
@@ -97,8 +78,8 @@ static void set_focus_app (ShellWindowTracker  *tracker,
                            ShellApp            *new_focus_app);
 static void on_focus_window_changed (MetaDisplay *display, GParamSpec *spec, ShellWindowTracker *tracker);
 
-static void track_window (ShellWindowTracker *monitor, MetaWindow *window);
-static void disassociate_window (ShellWindowTracker *monitor, MetaWindow *window);
+static void track_window (ShellWindowTracker *tracker, MetaWindow *window);
+static void disassociate_window (ShellWindowTracker *tracker, MetaWindow *window);
 
 
 static void
@@ -160,47 +141,6 @@ shell_window_tracker_class_init (ShellWindowTrackerClass *klass)
                                                    G_TYPE_NONE, 0);
 }
 
-/**
- * get_app_id_from_title:
- *
- * Use a window's "title" property to determine an application ID.
- * This is a temporary crutch for a few applications until we get
- * them correctly setting their WM_CLASS.
- */
-static const char *
-get_app_id_from_title (MetaWindow   *window)
-{
-  static gboolean patterns_initialized = FALSE;
-  const char *title;
-  int i;
-
-  title = meta_window_get_title (window);
-
-  if (!patterns_initialized) /* Generate match patterns once for all */
-    {
-      patterns_initialized = TRUE;
-      for (i = 0; title_patterns[i].app_id; i++)
-        {
-          title_patterns[i].regex = g_regex_new (title_patterns[i].pattern,
-                                                 0, 0, NULL);
-        }
-    }
-
-  /* Match window title patterns to identifiers for non-standard apps */
-  if (title)
-    {
-      for (i = 0; title_patterns[i].app_id; i++)
-        {
-          if (g_regex_match (title_patterns[i].regex, title, 0, NULL))
-            {
-              /* Matched, return the app id we want */
-              return title_patterns[i].app_id;
-            }
-        }
-    }
-  return NULL;
-}
-
 /**
  * get_appid_from_window:
  *
@@ -303,16 +243,10 @@ get_app_from_window_wmclass (MetaWindow  *window)
   g_free (wmclass);
 
   app = shell_app_system_lookup_heuristic_basename (appsys, with_desktop);
+  if (app != NULL)
+    g_object_ref (app);
   g_free (with_desktop);
 
-  if (app == NULL)
-    {
-      const char *id = get_app_id_from_title (window);
-
-      if (id != NULL)
-        app = shell_app_system_get_app (appsys, id);
-    }
-
   return app;
 }
 
@@ -327,7 +261,7 @@ get_app_from_window_wmclass (MetaWindow  *window)
  * Return value: (transfer full): A newly-referenced #ShellApp, or %NULL
  */
 static ShellApp*
-get_app_from_window_group (ShellWindowTracker  *monitor,
+get_app_from_window_group (ShellWindowTracker  *tracker,
                            MetaWindow          *window)
 {
   ShellApp *result;
@@ -351,7 +285,7 @@ get_app_from_window_group (ShellWindowTracker  *monitor,
       if (meta_window_get_window_type (group_window) != META_WINDOW_NORMAL)
         continue;
 
-      result = g_hash_table_lookup (monitor->window_to_app, group_window);
+      result = g_hash_table_lookup (tracker->window_to_app, group_window);
       if (result)
         break;
     }
@@ -366,7 +300,7 @@ get_app_from_window_group (ShellWindowTracker  *monitor,
 
 /**
  * get_app_from_window_pid:
- * @monitor: a #ShellWindowTracker
+ * @tracker: a #ShellWindowTracker
  * @window: a #MetaWindow
  *
  * Check if the pid associated with @window corresponds to an
@@ -406,22 +340,19 @@ get_app_from_window_pid (ShellWindowTracker  *tracker,
  * Returns: (transfer full): a #ShellApp, or NULL if none is found
  */
 static ShellApp *
-get_app_for_window (ShellWindowTracker    *monitor,
-                    MetaWindow         *window)
+get_app_for_window (ShellWindowTracker    *tracker,
+                    MetaWindow            *window)
 {
-  ShellApp *result;
+  ShellApp *result = NULL;
   const char *startup_id;
 
-  if (meta_window_is_remote (window))
-    return shell_app_system_get_app_for_window (shell_app_system_get_default (), window);
-
-  result = NULL;
   /* First, we check whether we already know about this window,
    * if so, just return that.
    */
-  if (meta_window_get_window_type (window) == META_WINDOW_NORMAL)
+  if (meta_window_get_window_type (window) == META_WINDOW_NORMAL
+      || meta_window_is_remote (window))
     {
-      result = g_hash_table_lookup (monitor->window_to_app, window);
+      result = g_hash_table_lookup (tracker->window_to_app, window);
       if (result != NULL)
         {
           g_object_ref (result);
@@ -429,6 +360,9 @@ get_app_for_window (ShellWindowTracker    *monitor,
         }
     }
 
+  if (meta_window_is_remote (window))
+    return _shell_app_new_for_window (window);
+
   /* Check if the app's WM_CLASS specifies an app; this is
    * canonical if it does.
    */
@@ -436,7 +370,7 @@ get_app_for_window (ShellWindowTracker    *monitor,
   if (result != NULL)
     return result;
 
-  result = get_app_from_window_pid (monitor, window);
+  result = get_app_from_window_pid (tracker, window);
   if (result != NULL)
     return result;
 
@@ -446,7 +380,7 @@ get_app_for_window (ShellWindowTracker    *monitor,
     {
       GSList *iter, *sequences;
 
-      sequences = shell_window_tracker_get_startup_sequences (monitor);
+      sequences = shell_window_tracker_get_startup_sequences (tracker);
       for (iter = sequences; iter; iter = iter->next)
         {
           ShellStartupSequence *sequence = iter->data;
@@ -456,7 +390,10 @@ get_app_for_window (ShellWindowTracker    *monitor,
 
           result = shell_startup_sequence_get_app (sequence);
           if (result)
-            break;
+            {
+              result = g_object_ref (result);
+              break;
+            }
         }
     }
 
@@ -464,11 +401,11 @@ get_app_for_window (ShellWindowTracker    *monitor,
    * any other windows in the group.
    */
   if (result == NULL)
-    result = get_app_from_window_group (monitor, window);
+    result = get_app_from_window_group (tracker, window);
 
   /* Our last resort - we create a fake app from the window */
   if (result == NULL)
-    result = shell_app_system_get_app_for_window (shell_app_system_get_default (), window);
+    result = _shell_app_new_for_window (window);
 
   return result;
 }
@@ -480,32 +417,30 @@ _shell_window_tracker_get_app_context (ShellWindowTracker *tracker, ShellApp *ap
 }
 
 static void
-on_transient_window_title_changed (MetaWindow      *window,
-                                   GParamSpec      *spec,
-                                   ShellWindowTracker *self)
+update_focus_app (ShellWindowTracker *self)
 {
-  ShellAppSystem *appsys;
-  ShellApp *app;
-  const char *id;
+  MetaWindow *new_focus_win;
+  ShellApp *new_focus_app;
 
-  /* Check if we now have a mapping using the window title */
-  id = get_app_id_from_title (window);
-  if (id == NULL)
-    return;
+  new_focus_win = meta_display_get_focus_window (shell_global_get_display (shell_global_get ()));
+  new_focus_app = new_focus_win ? shell_window_tracker_get_window_app (self, new_focus_win) : NULL;
 
-  appsys = shell_app_system_get_default ();
-  app = shell_app_system_get_app (appsys, id);
-  if (app == NULL)
-    return;
-  g_object_unref (app);
+  set_focus_app (self, new_focus_app);
+}
 
-  /* We found an app, don't listen for further title changes */
-  g_signal_handlers_disconnect_by_func (window, G_CALLBACK (on_transient_window_title_changed),
-                                        self);
+static void
+on_wm_class_changed (MetaWindow  *window,
+                     GParamSpec  *pspec,
+                     gpointer     user_data)
+{
+  ShellWindowTracker *self = SHELL_WINDOW_TRACKER (user_data);
 
   /* It's simplest to just treat this as a remove + add. */
   disassociate_window (self, window);
   track_window (self, window);
+  /* also just recaulcuate the focused app, in case it was the focused
+     window that changed */
+  update_focus_app (self);
 }
 
 static void
@@ -524,15 +459,7 @@ track_window (ShellWindowTracker *self,
   /* At this point we've stored the association from window -> application */
   g_hash_table_insert (self->window_to_app, window, app);
 
-  if (shell_app_is_transient (app))
-    {
-      /* For a transient application, it's possible one of our title regexps
-       * will match at a later time, i.e. the application may not have set
-       * its title fully at the time it initially maps a window.  Watch
-       * for title changes and recompute the app.
-       */
-      g_signal_connect (window, "notify::title", G_CALLBACK (on_transient_window_title_changed), self);
-    }
+  g_signal_connect (window, "notify::wm-class", G_CALLBACK (on_wm_class_changed), self);
 
   _shell_app_add_window (app, window);
 
@@ -564,7 +491,10 @@ disassociate_window (ShellWindowTracker   *self,
   g_hash_table_remove (self->window_to_app, window);
 
   if (shell_window_tracker_is_window_interesting (window))
+    {
       _shell_app_remove_window (app, window);
+      g_signal_handlers_disconnect_by_func (window, G_CALLBACK(on_wm_class_changed), self);
+    }
 
   g_signal_emit (self, signals[TRACKED_WINDOWS_CHANGED], 0);
 
@@ -580,7 +510,7 @@ shell_window_tracker_on_window_removed (MetaWorkspace   *workspace,
 }
 
 static void
-load_initial_windows (ShellWindowTracker *monitor)
+load_initial_windows (ShellWindowTracker *tracker)
 {
   GList *workspaces, *iter;
   MetaScreen *screen = shell_global_get_screen (shell_global_get ());
@@ -595,7 +525,7 @@ load_initial_windows (ShellWindowTracker *monitor)
       for (window_iter = windows; window_iter; window_iter = window_iter->next)
         {
           MetaWindow *window = window_iter->data;
-          track_window (monitor, window);
+          track_window (tracker, window);
         }
 
       g_list_free (windows);
@@ -709,28 +639,24 @@ static void
 shell_window_tracker_finalize (GObject *object)
 {
   ShellWindowTracker *self = SHELL_WINDOW_TRACKER (object);
-  int i;
 
   g_hash_table_destroy (self->running_apps);
   g_hash_table_destroy (self->window_to_app);
   g_hash_table_destroy (self->launched_pid_to_app);
 
-  for (i = 0; title_patterns[i].app_id; i++)
-    g_regex_unref (title_patterns[i].regex);
-
   G_OBJECT_CLASS (shell_window_tracker_parent_class)->finalize(object);
 }
 
 /**
  * shell_window_tracker_get_window_app:
- * @monitor: An app monitor instance
+ * @tracker: An app monitor instance
  * @metawin: A #MetaWindow
  *
  * Returns: (transfer full): Application associated with window
  */
 ShellApp *
-shell_window_tracker_get_window_app (ShellWindowTracker *monitor,
-                                  MetaWindow      *metawin)
+shell_window_tracker_get_window_app (ShellWindowTracker *tracker,
+                                     MetaWindow         *metawin)
 {
   MetaWindow *transient_for;
   ShellApp *app;
@@ -739,7 +665,7 @@ shell_window_tracker_get_window_app (ShellWindowTracker *monitor,
   if (transient_for != NULL)
     metawin = transient_for;
 
-  app = g_hash_table_lookup (monitor->window_to_app, metawin);
+  app = g_hash_table_lookup (tracker->window_to_app, metawin);
   if (app)
     g_object_ref (app);
 
@@ -792,7 +718,7 @@ shell_window_tracker_get_app_from_pid (ShellWindowTracker *self,
 
 /**
  * shell_window_tracker_get_running_apps:
- * @monitor: An app monitor instance
+ * @tracker: An app monitor instance
  * @context: Activity identifier
  *
  * Returns the set of applications which currently have at least one open
@@ -802,21 +728,21 @@ shell_window_tracker_get_app_from_pid (ShellWindowTracker *self,
  * Returns: (element-type ShellApp) (transfer full): Active applications
  */
 GSList *
-shell_window_tracker_get_running_apps (ShellWindowTracker *monitor,
-                                    const char      *context)
+shell_window_tracker_get_running_apps (ShellWindowTracker *tracker,
+                                       const char         *context)
 {
   gpointer key, value;
   GSList *ret;
   GHashTableIter iter;
 
-  g_hash_table_iter_init (&iter, monitor->running_apps);
+  g_hash_table_iter_init (&iter, tracker->running_apps);
 
   ret = NULL;
   while (g_hash_table_iter_next (&iter, &key, &value))
     {
       ShellApp *app = value;
 
-      if (strcmp (context, _shell_window_tracker_get_app_context (monitor, app)) != 0)
+      if (strcmp (context, _shell_window_tracker_get_app_context (tracker, app)) != 0)
         continue;
 
       ret = g_slist_prepend (ret, g_object_ref (app));
@@ -884,13 +810,7 @@ on_focus_window_changed (MetaDisplay        *display,
                          GParamSpec         *spec,
                          ShellWindowTracker *tracker)
 {
-  MetaWindow *new_focus_win;
-  ShellApp *new_focus_app;
-
-  new_focus_win = meta_display_get_focus_window (display);
-  new_focus_app = new_focus_win ? shell_window_tracker_get_window_app (tracker, new_focus_win) : NULL;
-
-  set_focus_app (tracker, new_focus_app);
+  update_focus_app (tracker);
 }
 
 /**
@@ -939,12 +859,11 @@ shell_startup_sequence_get_id (ShellStartupSequence *sequence)
  * shell_startup_sequence_get_app:
  * @sequence: A #ShellStartupSequence
  *
- * Returns: (transfer full): The application being launched, or %NULL if unknown.
+ * Returns: (transfer none): The application being launched, or %NULL if unknown.
  */
 ShellApp *
 shell_startup_sequence_get_app (ShellStartupSequence *sequence)
 {
-#ifdef HAVE_SN_STARTUP_SEQUENCE_GET_APPLICATION_ID
   const char *appid;
   ShellAppSystem *appsys;
   ShellApp *app;
@@ -954,11 +873,8 @@ shell_startup_sequence_get_app (ShellStartupSequence *sequence)
     return NULL;
 
   appsys = shell_app_system_get_default ();
-  app = shell_app_system_get_app_for_path (appsys, appid);
+  app = shell_app_system_lookup_app_for_path (appsys, appid);
   return app;
-#else
-  return NULL;
-#endif
 }
 
 const char *
